#!/usr/bin/env python3
"""
An abstract class for plot classes, can plot XY scatter/line (pending self.formatString) plot.
    If you specify --outputFname, make sure its suffix is .png.
    If "-i ..." is given, it is regarded as one of the input files 
        (plus the ones in trailing arguments). 

Examples:
    # Draw data column NumberOfLoci (y-axis, -W) vs. AAC (x-axis -l ...).
    # sample all data (-s 1), generate svg figure (-n)
    # take positive (-p) log (-g) of the whichColumn value (y-axis)
    %s -p -g -l AAC -W NumberOfLoci -D NoOfLoci -O ~/NoOfLoci_vs_AAC -n -s 1 -x AAC
        VCFStat_Method8_L800000P4000000m1000000.2012.8.1T0331/11Contigs_AAC_tally.tsv

"""
import sys, os, math
__doc__ = __doc__%(sys.argv[0])

import matplotlib; matplotlib.use("Agg")    #to disable pop-up requirement
import pylab
import logging
import statsmodels.api as sm
from palos import ProcessOptions, PassingData
from palos.plot import yh_matplotlib
from palos.io.AbstractMatrixFileWalker import AbstractMatrixFileWalker

class AbstractPlot(AbstractMatrixFileWalker):
    __doc__ = __doc__
    option_default_dict = AbstractMatrixFileWalker.option_default_dict.copy()
    option_default_dict.update({

    ('title', 0, ): [None, 't', 1, 'title for the figure.'],\
    ('figureDPI', 1, int): [200, 'f', 1, 
        'dpi, dots per inch, for the output figures (png)'],\
    ('formatString', 1, ): ['.', '', 1, 'formatString passed to matplotlib plot'],
    ('markerSize', 1, int): [10, '', 1, \
        'size of plotting marker (dot size, usually), matplotlib default is 5'],\
    ('ylim_type', 1, int): [1, 'y', 1, \
        'y-axis limit type, 1: whatever matplotlib decides. 2: min to max'],\
    ('xColumnHeader', 1, ): ['', 'l', 1, 'header of the x-axis data column, ' ],
    ('xColumnPlotLabel', 0, ): ['', 'x', 1,
        'x-axis label (posColumn) in manhattan plot', ],
    ('whichColumnPlotLabel', 0, ): ['', 'D', 1,
        'plot label for data of the whichColumn', ],
    ('fitType', 0, int): [0, '', 1, 'To add a fitting line. 0: nothing; 1: lowess/loess'],\

    ('logX', 0, int): [0, '', 1, 'To change X values. 0: nothing; 1: log(X), 2: -log(X)'],\
    ('xScaleLog', 0, int): [0, '', 1, \
        'regarding the x-axis scale. 0: nothing; 1: scale of log(10), 2: scale of log(2), '
        'if this is non-zero, then logX should be 0. Otherwise data will be messed up.'],\
    ('yScaleLog', 0, int): [0, '', 1, \
        'regarding the y-axis scale. 0: nothing; 1: scale of log(10), 2: scale of log(2), '
        'if this is non-zero, then logY should be 0. Otherwise data will be messed up.'],\
    
    ('defaultFontLabelSize', 1, int): [12, '', 1, 'default font & label size on the plot'], \
    ('defaultFigureWidth', 1, float): [12, '', 1, 
        'default figure width (in inch)'], \
    ('defaultFigureHeight', 1, float): [8, '', 1, \
        'default figure height (in inch)'], \
    ('xmargin', 0, float): [0, '', 1, \
        'this is margin within the plotting area.'
        'how far data points should be away from axis. 0-1.0'], \
    ('ymargin', 0, float): [0, '', 1, \
        'this is margin within the plotting area. '
        'how far data points should be away from axis. 0-1.0'], \
    ('plotLeft', 0, float): [0.1, '', 1, 'the left side of the subplots of the figure'], \
    ('plotRight', 0, float): [0.9, '', 1, 'the right side of the subplots of the figure'], \
    ('plotBottom', 0, float): [0.1, '', 1, ' the bottom of the subplots of the figure'], \
    ('plotTop', 0, float): [0.9, '', 1, ' the top of the subplots of the figure'], \
    ('legendType', 0, int): [0, '', 1, '0: no legend; 1: legend'], \
    ('need_svg', 0, ): [0, 'n', 0, 'whether need svg output', ],\
                        
    })
    def __init__(self, inputFnameLs=None, **keywords):
        """
        """
        #self.connectDB() called within its __init__()
        AbstractMatrixFileWalker.__init__(self, inputFnameLs=inputFnameLs, **keywords)
        #if user wants to preserve data in a data structure that is visisble
                # throughout reading different files.
        # then use this self.invariantPData.
        #AbstractMatrixFileWalker has initialized a structure like below.
        #self.invariantPData = PassingData()
        self.xMin = None
        self.xMax = None
        self.yMin = None
        self.yMax = None
        self.plotObjectLs = []
        self.plotObjectLegendLs = []
        
        # set these attributes if the class or its descendants have valid values
        if getattr(self, 'defaultFontLabelSize', None):
            yh_matplotlib.setFontAndLabelSize(self.defaultFontLabelSize)
        if getattr(self, "defaultFigureWidth", None) and \
                getattr(self, "defaultFigureHeight", None):
            yh_matplotlib.setDefaultFigureSize((self.defaultFigureWidth, 
                self.defaultFigureHeight))
        if getattr(self, "plotLeft", None) and getattr(self, "plotBottom", None):
            yh_matplotlib.setPlotDimension(left=self.plotLeft, right=self.plotRight,
                bottom=self.plotBottom, top=self.plotTop)

    def addPlotLegend(self, plotObject=None, legend=None, pdata=None, **keywords):
        """
        """
        self.plotObjectLs.append(plotObject)
        self.plotObjectLegendLs.append(legend)
    
    def plot(self, x_ls=None, y_ls=None, pdata=None, **keywords):
        """
        get called by the end of fileWalker() for each inputFname.
        """
        plotObject = pylab.plot(x_ls, y_ls, self.formatString, \
            markersize=self.markerSize)[0]
        self.addPlotLegend(plotObject=plotObject,
            legend=os.path.basename(pdata.filename), pdata=pdata)
        
        self.setGlobalMinVariable(extremeVariableName='xMin', givenExtremeValue=min(x_ls))
        self.setGlobalMaxVariable(extremeVariableName='xMax', givenExtremeValue=max(x_ls))
        self.setGlobalMinVariable(extremeVariableName='yMin', givenExtremeValue=min(y_ls))
        self.setGlobalMaxVariable(extremeVariableName='yMax', givenExtremeValue=max(y_ls))

        if self.fitType==1:
            # lowess will return our "smoothed" data with a y value for at every x-value
            lowess = sm.nonparametric.lowess(y_ls, x_ls, frac=.3)

            # unpack the lowess smoothed points to their values
            lowess_x, lowess_y = list(zip(*lowess))
            lowess_obj = pylab.plot(lowess_x, lowess_y, '.-')[0]
            self.addPlotLegend(plotObject=lowess_obj,
                legend="lowess", pdata=pdata)
    
    def processRow(self, row=None, pdata=None):
        """
        Deal with missing data.
        Handle one row a time.
        """
        col_name2index = getattr(pdata, 'col_name2index', None)
        x_ls = getattr(pdata, 'x_ls', None)
        y_ls = getattr(pdata, 'y_ls', None)
        if col_name2index and x_ls is not None and y_ls is not None:
            if self.whichColumnHeader:
                whichColumn = col_name2index.get(self.whichColumnHeader, None)
            else:
                whichColumn = self.whichColumn
            x_index = col_name2index.get(self.xColumnHeader, None)
            
            xValue = row[x_index]
            yValue = row[whichColumn]
            if xValue and yValue and yValue not in self.missingDataNotation and \
                    xValue not in self.missingDataNotation:
                xValue = self.processValue(xValue, processType=self.logX,
                    valueForNonPositiveValue=self.valueForNonPositiveYValue)
                yValue = self.processValue(yValue, processType=self.logY,
                    valueForNonPositiveValue=self.valueForNonPositiveYValue)
                x_ls.append(xValue)
                y_ls.append(yValue)
    
    def setGlobalExtremeVariable(self, extremeVariableName=None,
        givenExtremeValue=None, extremeType=1):
        """
        extremeType:
            1: min
            2: max
        """
        if givenExtremeValue is not None:
            extremeVariable = getattr(self, extremeVariableName)
            if extremeVariable is None:
                setattr(self, extremeVariableName, givenExtremeValue)
            else:
                if extremeType==1:    #replace a min variable
                    if givenExtremeValue<extremeVariable:
                        setattr(self, extremeVariableName, givenExtremeValue)
                else:    #set a max variable
                    if givenExtremeValue>extremeVariable:
                        setattr(self, extremeVariableName, givenExtremeValue)
    
    def setGlobalMinVariable(self, extremeVariableName=None, givenExtremeValue=None):
        """
        """
        self.setGlobalExtremeVariable(extremeVariableName=extremeVariableName,
                    givenExtremeValue=givenExtremeValue, \
            extremeType=1)
    
    def setGlobalMaxVariable(self, extremeVariableName=None, givenExtremeValue=None):
        """
        """
        self.setGlobalExtremeVariable(extremeVariableName=extremeVariableName, 
            givenExtremeValue=givenExtremeValue, extremeType=2)
        
    def processHeader(self, header=None, pdata=None):
        """
        called everytime the header of an input file is derived in fileWalker()
        """
        pass
    
    def _handleXLim(self, **keywords):
        """
        Could be used by descendants, but not called by default
        """
        if self.xMin is not None and self.xMax is not None:
            delta = abs(self.xMax-self.xMin)/10.0
            if delta<=0:
                delta = 0.5
            pylab.xlim(xmin=self.xMin-delta, xmax=self.xMax+delta)
        
    def _handleYLim(self, **keywords):
        """
        Could be used by descendants, but not called by default
        """
        if self.yMin is not None and self.yMax is not None:
            #originalYMin, originalYMax = pylab.ylim()
            delta = abs(self.yMax-self.yMin)/10.0
            if delta<=0:
                delta = 0.5
            pylab.ylim(ymin=self.yMin-delta, ymax=self.yMax+delta)
    
    def handleXLim(self, **keywords):
        """
        """
        pass
        
    def handleYLim(self, **keywords):
        """
        """
        pass
    
    def handleXLabel(self, **keywords):
        """
        """
        if getattr(self, 'xColumnPlotLabel', None):
            xlabel = self.xColumnPlotLabel
        else:
            xlabel = getattr(self, "xColumnHeader", "")
        pylab.xlabel(xlabel)
        return xlabel
        
    def handleYLabel(self, **keywords):
        """
        """
        if getattr(self, 'whichColumnPlotLabel', None):
            ylabel = self.whichColumnPlotLabel
        else:
            ylabel = getattr(self, "whichColumnHeader", "")
        pylab.ylabel(ylabel)
        return ylabel
    
    def handleTitle(self, **keywords):
        """
        """
        if self.title:
            title = self.title
        else:
            title = yh_matplotlib.constructTitleFromTwoDataSummaryStat(
                self.invariantPData.x_ls, self.invariantPData.y_ls)
        pylab.title(title)
        return title
    
    def changeFigureScaleToLog(self, xScaleLog=None, yScaleLog=None):
        """
        Not used in default mode
        """
        
        if xScaleLog is None:
            xScaleLog = self.xScaleLog
        if yScaleLog is None:
            yScaleLog = self.yScaleLog
        if xScaleLog==1:
            #only change the axis into log-scale if it's positive log
            pylab.gca().set_xscale('log', basex=10)
        elif xScaleLog==2:
            pylab.gca().set_xscale('log', basex=2)
        
        if yScaleLog==1:
            #only change the axis into log-scale if it's positive log
            pylab.gca().set_yscale('log', basey=10)
        elif yScaleLog==2:
            pylab.gca().set_yscale('log', basey=2)
        
    def saveFigure(self, invariantPData=None, **keywords):
        """
        add legend or not , depending on self.legendType
        """
        sys.stderr.write("Saving figure ")
        # change scale
        self.changeFigureScaleToLog()
        # change margin
        self.setMargins()
        
        if self.legendType!=0 and self.plotObjectLegendLs and self.plotObjectLs:
            #add the legend
            pylab.legend(self.plotObjectLs, self.plotObjectLegendLs, shadow = True)
        
        if self.outputFnamePrefix:
            pngOutputFname = '%s.png'%self.outputFnamePrefix
            svgOutputFname = '%s.svg'%self.outputFnamePrefix
        elif self.outputFname:
            pngOutputFname = self.outputFname
            svgOutputFname = '%s.svg'%(self.outputFname[:-4])
        else:
            logging.error(f"Could not get outputFnamePrefix from "
                f"self.outputFnamePrefix {self.outputFnamePrefix} or "
                f"self.outputFname {self.outputFname}.")
            sys.exit(1)
        sys.stderr.write("to %s ..."%(pngOutputFname))
        pylab.savefig(pngOutputFname, dpi=self.figureDPI)
        if self.need_svg:
            pylab.savefig(svgOutputFname, dpi=self.figureDPI)
        sys.stderr.write("  .\n")

    def setup(self, **keywords):
        """
        run before anything is run
        """
        AbstractMatrixFileWalker.setup(self, **keywords)
        pylab.clf()
    
    def reduce(self, **keywords):
        """
        run after all files have been walked through
        """
        self.saveFigure(invariantPData=self.invariantPData)
        #delete self.invariantPData.writer if it exists
        AbstractMatrixFileWalker.reduce(self, **keywords)
    
    def setMargins(self):
        """
        """
        if self.xmargin is not None:
            pylab.gca().set_xmargin(self.xmargin)
        if self.ymargin is not None:
            pylab.gca().set_ymargin(self.ymargin)
    
    def run(self):
        
        if self.debug:
            import pdb
            pdb.set_trace()
        
        self.setup()
        
        for inputFname in self.inputFnameLs:
            if os.path.isfile(inputFname):
                self.fileWalker(inputFname, afterFileFunction=None, 
                    run_type=1, processRowFunction=self.processRow)
                #afterFileFunction = None means self.plot
        
        self.handleTitle()
        self.handleXLabel()
        self.handleYLabel()
        self.handleXLim()
        self.handleYLim()
        self.reduce()
        
        self.closeFiles()
        

if __name__ == '__main__':
    main_class = AbstractPlot
    po = ProcessOptions(sys.argv, main_class.option_default_dict,
        error_doc=main_class.__doc__)
    instance = main_class(po.arguments, **po.long_option2value)
    instance.run()



#!/usr/bin/env python3
"""
2008-02-02 extend the artists in matplotlib
"""
import math
from numpy import array, arange, sin, cos, pi, sqrt, sqrt, nonzero, equal, asarray, dot, concatenate
import numpy
#2012.7.9 replace matplotlib.numerix with numpy. the former is gone.
from matplotlib.artist import Artist, setp, kwdoc
from matplotlib.cbook import dedent
from matplotlib.patches import Polygon
from matplotlib.collections import PolyCollection, LineCollection

# these are not available for the object inspector until after the
# class is build so we define an initial set here for the init
# function and they will be overridden after object defn
kwdocd = kwdoc.func_dict	#2012.7.9 kwdocd is no longer available from matplotlib.artist
kwdocd['Patch'] = """\
		  alpha: float
		  animated: [True | False]
		  antialiased or aa: [True | False]
		  clip_box: a matplotlib.transform.Bbox instance
		  clip_on: [True | False]
		  edgecolor or ec: any matplotlib color
		  facecolor or fc: any matplotlib color
		  figure: a matplotlib.figure.Figure instance
		  fill: [True | False]
		  hatch: unknown
		  label: any string
		  linewidth or lw: float
		  lod: [True | False]
		  transform: a matplotlib.transform transformation instance
		  visible: [True | False]
		  zorder: any number
		  """

class Gene(Polygon):
	"""2008-02-04 An artist for gene. block-like. based on matplotlib.patches.FancyArrow"""

	def __init__(self, start_ls, end_ls, y=0, width=0.001, x_offset=0, is_arrow=True, length_includes_head=True, \
		head_width=None, head_length=None, shape='full', overhang=0, \
		head_starts_at_zero=False,**kwargs):
		"""2008-02-02 an artist for gene. block-like. based on matplotlib.patches.FancyArrow
		
		Returns a new Arrow.
		
		x_offset: is the value of how far start_ls and end_ls should all be pushed
		
		is_arrow: True if it's an arrow. False if it's just a rectangle
		
		length_includes_head: True if head is counted in calculating the length.

		shape: ['full', 'left', 'right']

		overhang: ratio of the head_length that the arrow is swept back (0 overhang means
		triangular shape).

		head_starts_at_zero: if True, the head starts being drawn at coordinate
		0 instead of ending at coordinate 0.

		Valid kwargs are:
		%(Patch)s

		"""
		if head_width is None:
			head_width = 2 * width
		if head_length is None:
			head_length = 1/(2.0 * abs(end_ls[0]-start_ls[0]))

		distance = abs(end_ls[-1]-start_ls[0])
		if length_includes_head:
			length=distance
		else:
			length=distance+head_length
		
		no_of_blocks = len(start_ls)
		if not distance:
			verts = [] #display nothing if empty
		else:
			"""
			start by drawing horizontal arrow, point (tip of the arrow) at (0,0). the whole arrow sticks on the x-axis (<=0 part)
			Notice: the XY -coordination is not the usual math coordination system. it's canvas coordination. (0,0) is top left. horizontal is y-axis. vertical is x-axis.
			start from the last block, reversely to the 1st block
			"""
			inc_block_length = abs(end_ls[-1]-start_ls[-1])
			hw, hl = head_width, head_length
			if is_arrow:
				left_half_arrow_ls = [
    				[0.0,0.0],				  #tip
    				[-hl, -hw/2.0],			 #leftmost
    				[-hl*(1-overhang), -width/2.0], #meets stem
    				[-inc_block_length, -width/2.0]		 #bottom left
    			]
			else:	#this is just a rectangle
				left_half_arrow_ls = [
    				[0.0,0.0],				  #tip
    				[0.0, -width/2.0],		#left to the tip
    				[-inc_block_length, -width/2.0]		 #bottom left
    			]
			if no_of_blocks==1:	#only one block, seal it
				left_half_arrow_ls.append([-inc_block_length, 0])
			elif no_of_blocks>1:	#more than one block
				left_half_arrow_ls.append([-inc_block_length, -width/6.0])   #leave if open
				for i in range(1, no_of_blocks):
					block_index = -i-1	#backwards
					gap = abs(start_ls[block_index+1]-end_ls[block_index])
					inc_block_start = inc_block_length + gap
					this_block_length = abs(end_ls[block_index]-start_ls[block_index])
					inc_block_length = inc_block_length + gap + this_block_length
					if i!=no_of_blocks-1:	#don't seal it
						left_half_arrow_ls.append([-inc_block_start, -width/6.0])
						left_half_arrow_ls.append([-inc_block_start, -width/2.0])
						left_half_arrow_ls.append([-inc_block_length, -width/2.0])
						left_half_arrow_ls.append([-inc_block_length, -width/6.0])
					else:	#seal it
						left_half_arrow_ls.append([-inc_block_start, -width/6.0])
						left_half_arrow_ls.append([-inc_block_start, -width/2.0])
						left_half_arrow_ls.append([-inc_block_length, -width/2.0])
						left_half_arrow_ls.append([-inc_block_length, 0])
			left_half_arrow = array(left_half_arrow_ls)
			#if we're not including the head, shift up by head length
			if not length_includes_head:
				left_half_arrow += [head_length, 0]
			#if the head starts at 0, shift up by another head length
			if head_starts_at_zero:
				left_half_arrow += [head_length/2.0, 0]
			#figure out the shape, and complete accordingly
			if shape == 'left':
				coords = left_half_arrow
			else:
				right_half_arrow = left_half_arrow*[1,-1]
				if shape == 'right':
					coords = right_half_arrow
				elif shape == 'full':
					coords=concatenate([left_half_arrow,right_half_arrow[::-1]])
				else:
					raise ValueError("Got unknown shape: %s" % shape)
			dx = end_ls[-1]-start_ls[0]
			x = start_ls[0] + x_offset
			dy = 0
			cx = float(dx)/distance
			sx = float(dy)/distance
			M = array([[cx, sx],[-sx,cx]])
			#verts = matrixmultiply(coords, M) + (x+dx, y+dy)
			verts =coords*M + (x+dx, y+dy)	#2012.7.9 replace matrixmultiply with "*" as matrixmultiply is not available from numpy
		Polygon.__init__(self, map(tuple, verts), **kwargs)
	__init__.__doc__ = dedent(__init__.__doc__) % kwdocd

class ExonCollection(PolyCollection):
	"""
	2008-09-23	test PolyCollection, only draw exons
	"""
	def __init__(self, start_ls, end_ls, y=0, width=0.001, x_offset=0, is_arrow=True, length_includes_head=True, \
		head_width=None, head_length=None, shape='full', overhang=0, \
		head_starts_at_zero=False,**kwargs):
		if head_width is None:
			head_width = 2 * width
		if head_length is None:
			head_length = 1/(2.0 * abs(end_ls[0]-start_ls[0]))

		distance = abs(end_ls[-1]-start_ls[0])
		if length_includes_head:
			length=distance
		else:
			length=distance+head_length
		
		no_of_blocks = len(start_ls)
		if not distance:
			verts_ls = [] #display nothing if empty
		else:
			"""
			start by drawing horizontal arrow, point (tip of the arrow) at (0,0). the whole arrow sticks on the x-axis (<=0 part)
			Notice: the XY -coordination is not the usual math coordination system. it's canvas coordination. (0,0) is top left. horizontal is y-axis. vertical is x-axis.
			start from the last block, reversely to the 1st block
			"""
			verts_ls = []
			for i in range(no_of_blocks):
				block_start = start_ls[i]
				block_end = end_ls[i]
				verts = [[block_start, width/2.0+y], [block_start, -width/2.0+y], [block_end, -width/2.0+y], [block_end, width/2.0+y]]
				
				verts_ls.append(verts)
		PolyCollection.__init__(self, verts_ls, **kwargs)

import matplotlib as mpl
import matplotlib.colors as _colors # avoid conflict with kwarg
import matplotlib.transforms as transforms
import matplotlib.cbook as cbook
class ExonIntronCollection20081130(PolyCollection, LineCollection):
	"""
	2010-4-12
		It became outdated as matplotlib is upgraded. use the new ExonIntronCollection below.
	2008-11-30
		add option rotate_xy to specify an anti-clockwise-90-degree rotation.
		default rotate_xy=False, gene models sit on the y-axis, strand direction matching the x-axis.
		if rotate_xy=True, gene models sit on the x-axis, strand direction matching the y-axis.
	2008-10-01
		it now can handle picker_event and alpha on both boxes and lines
	2008-09-23	test PolyCollection, draw both exons and introns
	"""
	def __init__(self, annotated_box_ls, y=0, width=0.001, linewidths = None, box_line_widths = None,
				 colors	= None,
				 antialiaseds = None,
				 linestyle = 'solid',
				 offsets = None,
				 transOffset = None,#transforms.identity_transform(),
				 norm = None,
				 cmap = None,
				 picker=False, pickradius=5, alpha=None, strand=None, facecolors=None, edgecolors=None, rotate_xy=False, \
				 **kwargs):
		"""
		2008-10-01
			block_type='exon' has both facecolor and edgecolor red.
			block_type='intron' has no facecolor and black edgecolor.
			arguments edgecolors and facecolors in __init__() could overwrite the ones that generated based on the box_type.
		2008-10-01
			the annotated_box_ls is a list of (start, stop, box_type, is_translated, ...)
			box_type = 'exon' or 'intron' and is_translated= 0 or 1. They are optional. Default is box_type ='intron', is_translated=0.
		2008-09-26
			linewidths controls the width of lines connecting exons or the arrow line, could be a list of floats of just one float.
			box_line_widths controls the width of the boundaries of exon boxes
		"""
		#handle parameters for LineCollection
		if linewidths is None   :
			linewidths   = (mpl.rcParams['lines.linewidth'], )
		if box_line_widths is None   :
			box_line_widths   = (mpl.rcParams['lines.linewidth'], )
		if colors is None	   :
			colors	   = (mpl.rcParams['lines.color'],)
		if antialiaseds is None :
			antialiaseds = (mpl.rcParams['lines.antialiased'], )
		
		self._colors = [_colors.colorConverter.to_rgba(color, alpha) for color in colors]	#2008-10-26 use to_rgba() instead of to_rgba_list() because the latter is not present in matplotlib 0.98.3
		self._aa = self._get_value(antialiaseds)
		self._lw = self._get_value(linewidths)
		self.set_linestyle(linestyle)
		self._uniform_offsets = None
		if offsets is not None:
			offsets = asarray(offsets)
			if len(offsets.shape) == 1:
				offsets = offsets[numpy.newaxis,:]  # Make it Nx2.
		if transOffset is None:
			if offsets is not None:
				self._uniform_offsets = offsets
				offsets = None
			if hasattr(transforms, 'identity_transform'):
				transOffset = transforms.identity_transform()	#matplotlib 0.91.2
			elif hasattr(transforms, 'IdentityTransform'):
				transOffset = transforms.IdentityTransform()	#matplotlib 0.98.3
		self._offsets = offsets
		self._transOffset = transOffset
		self.pickradius = pickradius
		
		#handle parameters for PolyCollection
		verts_ls = []
		segment_ls = []
		no_of_blocks = len(annotated_box_ls)
		verts_colors = []
		_edgecolors = []
		for i in range(no_of_blocks):
			this_block = annotated_box_ls[i]
			block_start, block_end = this_block[:2]
			block_type = 'intron'
			is_translated = 0
			if len(this_block)>=3:
				block_type = this_block[2]
			if len(this_block)>=4:
				is_translated = int(this_block[3])
			if block_type=='exon':
				verts = [[block_start, width/2.0+y], [block_start, -width/2.0+y], [block_end, -width/2.0+y], [block_end, width/2.0+y]]
				if rotate_xy:
					verts = [[row[1], row[0]] for row in verts]
				verts_ls.append(verts)
				if is_translated==1:
					verts_colors.append('r')
					_edgecolors.append('r')	#same edge color
				else:
					verts_colors.append('w')	#no face color
					_edgecolors.append('k')
			else:
				if rotate_xy:
					segment_ls.append([(y, block_start), (y, block_end)])
				else:
					segment_ls.append([(block_start, y), (block_end,y)])
		if not facecolors:
			facecolors = verts_colors
		if not edgecolors:
			edgecolors = _edgecolors
		#add an arrow
		box_start = annotated_box_ls[0][0]
		box_end = annotated_box_ls[-1][1]
		if strand is not None:
			arrow_length = abs(box_start-box_end)/4
			try:
				strand = int(strand)
			except:
				pass
			if strand==1 or strand=='plus':
				arrow_offset = -arrow_length
				arrow_end = box_end
			elif strand==-1 or strand=='minus':
				arrow_offset = arrow_length
				arrow_end = box_start
			else:	#can't tell which strand
				arrow_offset = 0
				arrow_end = box_end
			if rotate_xy:
				segment_ls.append([( y+width, arrow_end+arrow_offset), (y+width/2.0, arrow_end)])
			else:
				segment_ls.append([(arrow_end+arrow_offset, y+width), (arrow_end,y+width/2.0)])
		self._segments = segment_ls
		self.set_segments(segment_ls)
		self._verts = verts
		#self.set_alpha(alpha)
		PolyCollection.__init__(self, verts_ls, linewidths=box_line_widths, antialiaseds=antialiaseds, edgecolors=edgecolors, **kwargs)
		self._facecolors  = [_colors.colorConverter.to_rgba(facecolor, alpha=alpha) for facecolor in facecolors]	#in PolyCollection.__init__(), it doesn't handle alpha
		self._picker = picker	#to enable picker_event. PolyCollection's ancestor PatchCollection (collection.py) defines a method 'contains()'.
			#Artist.pick() (in artist.py) calls pickable(), which checks self._picker, first to see whether this artist is pickable. then it calls picker() or contains().
		self.update(kwargs)
	
	def draw(self, renderer):
		if not self.get_visible(): return
		renderer.open_group('polycollection')
		transform = self.get_transform()
		if hasattr(self, 'get_transoffset'):
			transoffset = self.get_transoffset()	#matplotlib 0.91.2
		else:
			transoffset = getattr(self, '_transOffset', None)	#matplotlib 0.98.3
		
		if hasattr(transform, 'freeze'):
			transform.freeze()
		if transoffset and hasattr(transoffset, 'freeze'):
			transoffset.freeze()
		self.update_scalarmappable()
		if cbook.is_string_like(self._edgecolors) and self._edgecolors[:2] == 'No':
			self._linewidths = (0,)
			#self._edgecolors = self._facecolors
		renderer.draw_poly_collection(
			self._verts, transform, self.clipbox,
			self._facecolors, self._edgecolors,
			self._linewidths, self._antialiaseds,
			self._offsets,  transoffset)
		
		self.update_scalarmappable()
		#print 'calling renderer draw line collection'
		offsets = self._offsets
		renderer.draw_line_collection(
			self._segments, transform, self.clipbox,
			self._colors, self._lw, self._ls, self._aa, offsets,
			transoffset)
		
		if hasattr(transform, 'freeze'):
			transform.thaw()
		if transoffset and hasattr(transoffset, 'freeze'):
			transoffset.thaw()
		renderer.close_group('polycollection')

def on_canvas_pick(event):
	print(dir(event))
	import pdb
	pdb.set_trace()


from matplotlib.patches import PathPatch 
from matplotlib.path import Path

class ExonIntronCollection(PathPatch):
	"""
	2010-4-12
		ExonIntronCollection has been broke since the new matplotlib (0.99.0).
		The new one has the same interface as ExonIntronCollection
	"""
	def __init__(self, annotated_box_ls, y=0, width=0.001, linewidths = None, box_line_widths = None,
				 colors	= None,
				 antialiaseds = None,
				 linestyle = 'solid',
				 offsets = None,
				 transOffset = None,#transforms.identity_transform(),
				 norm = None,
				 cmap = None,
				 picker=False, pickradius=5, alpha=None, strand=None, facecolors=None, edgecolors=None, rotate_xy=False, \
				 **kwargs):
		"""
		2010-4-13
			the annotated_box_ls is a list of (start, stop, box_type, is_translated, ...)
			
			block_type='exon' has the whole "width" as its height
			block_type='intron' has width/2.0 as its height
			box_type = 'exon' or 'intron' and is_translated= 0 or 1. They are optional. Default is box_type ='intron', is_translated=0.
			
			linewidths controls the width of lines connecting exons or the arrow line, just one float number.
			
			argument rotate_xy to specify an anti-clockwise-90-degree rotation.
				default rotate_xy=False, gene models sit on the y-axis, strand direction matching the x-axis.
				if rotate_xy=True, gene models sit on the x-axis, strand direction matching the y-axis.
			if argument picker is True, it'll allow picker_event.
			argument alpha applies to both boxes and lines.
			
			box_line_widths is ignored.
			arguments edgecolors and facecolors in __init__() are ignored.
		"""
		if linewidths is None   :
			linewidths   = (mpl.rcParams['lines.linewidth'], )
		
		#handle parameters for PolyCollection
		vert_ls = []
		code_ls = []
		
		segment_ls = []
		no_of_blocks = len(annotated_box_ls)
		verts_colors = []
		_edgecolors = []
		for i in range(no_of_blocks):
			this_block = annotated_box_ls[i]
			block_start, block_end = this_block[:2]
			block_type = 'intron'
			is_translated = 0
			if len(this_block)>=3:
				block_type = this_block[2]
			if len(this_block)>=4:
				is_translated = int(this_block[3])
			if block_type=='exon':
				if is_translated==1:
					block_height = width/2.0
				else:
					block_height = width/4.0
				vert = [[block_start, block_height+y], [block_start, -block_height+y], [block_end, -block_height+y], [block_end, block_height+y], [block_start, block_height+y]]
				code = [Path.MOVETO, Path.LINETO, Path.LINETO, Path.LINETO, Path.CLOSEPOLY,]
				if rotate_xy:
					vert = [[row[1], row[0]] for row in vert]
				vert_ls.extend(vert)
				code_ls.extend(code)
			else:
				vert = [(block_start, y), (block_end,y)]
				if rotate_xy:
					vert = [[row[1], row[0]] for row in vert]
				code = [Path.MOVETO, Path.LINETO,]
				vert_ls.extend(vert)
				code_ls.extend(code)
		#add an arrow
		box_start = annotated_box_ls[0][0]
		box_end = annotated_box_ls[-1][1]
		if strand is not None:
			arrow_length = abs(box_start-box_end)/4
			try:
				strand = int(strand)
			except:
				pass
			if strand==1 or strand=='plus':
				arrow_offset = -arrow_length
				arrow_end = box_end
			elif strand==-1 or strand=='minus':
				arrow_offset = arrow_length
				arrow_end = box_start
			else:	#can't tell which strand
				arrow_offset = 0
				arrow_end = box_end
			vert = [(arrow_end+arrow_offset, y+width), (arrow_end, y+width), (arrow_end,y+width/4.0),(arrow_end,y+width/4.0)]
			if rotate_xy:
				vert = [[row[1], row[0]] for row in vert]
			code = [Path.MOVETO, Path.CURVE4, Path.CURVE4, Path.LINETO,]
			vert_ls.extend(vert)
			code_ls.extend(code)
		from matplotlib import path
		barpath = Path(vert_ls, code_ls)
		PathPatch.__init__(self, barpath, facecolor='red', edgecolor='gray', alpha=alpha, picker=picker, linewidth=linewidths)
		

if __name__ == '__main__':
	import pylab
	a = pylab.gca()
	fig = pylab.gcf()
	fig.canvas.mpl_connect('pick_event', on_canvas_pick)
	start_ls = [1,3,5]
	end_ls = [2,3.5,6]
	pylab.scatter(start_ls, end_ls)
	g = Gene(start_ls, end_ls, y=1,  width=0.1, alpha=0.3, facecolor='r', picker=True)
	a.add_artist(g)
	
	#draw a rectangle
	g1 = Gene(start_ls, end_ls, y=1.5,  width=0.1, x_offset=1, is_arrow=False, alpha=0.3, facecolor='r', picker=True)
	a.add_artist(g1)
	annotated_box_ls = [(1, 1.5, 'exon', 0),(1.5, 2, 'exon', 1), (2,3,'intron'), (3, 3.5, 'exon', 1), (3.5, 5, 'intron'), (5,6,'exon', 1)]
	g2 = ExonIntronCollection(annotated_box_ls, y=0.5, strand=1, width=0.1, facecolors=['y','w', 'w'], alpha=0.3, box_line_widths=0.3, picker=True)
	a.add_artist(g2)
	#a.add_patch(g2)
	#draw a opposite strand gene
	start_ls.reverse()
	end_ls.reverse()
	#g2 = ExonCollection(end_ls, start_ls, y=0.5,  width=0.1, x_offset=0.5, head_length=1, facecolors=['r','w', 'w'], overhang=0.3)
	g3 = ExonIntronCollection(annotated_box_ls, strand=-1, y=0,  width=0.1, facecolors=['g','w', 'w'], alpha=0.3, picker=True)
	a.add_patch(g3)
	a.set_xlim(min(start_ls), max(end_ls))
	#a.set_ylim(0,2)
	from matplotlib.patches import Circle, Polygon
	from matplotlib.lines import Line2D
	p1 = Polygon(zip(start_ls, end_ls), alpha=0.3, linewidth=0, facecolor=(0,217/255.,1), picker=True)
	a.add_artist(p1)
	#pylab.gray()	#colorbar needs  "You must first set_array for mappable"
	c1 = Circle((1,2), radius=1, facecolor='w', edgecolor='r')
	a.add_artist(c1)
	#pylab.savefig('abc.png', dpi=300)
	#pylab.savefig('abc.svg', dpi=300)
	pylab.show()



#!/usr/bin/env python3
"""
abstract class for pegasus workflows that work on alignment files (already aligned).
"""
import sys, os
import copy
import logging
from typing import List
from pegaflow.api import File, Job
from palos import ProcessOptions, PassingData
from . MapReduceGenomeFileWorkflow import MapReduceGenomeFileWorkflow

ParentClass = MapReduceGenomeFileWorkflow
class AbstractAlignmentWorkflow(ParentClass):
    __doc__ = __doc__
    option_default_dict = copy.deepcopy(ParentClass.option_default_dict)
    commonAlignmentWorkflowOptionDict = {
        ('ind_seq_id_ls', 0, ): ['', 'i', 1, 
            'a comma/dash-separated list of IndividualSequence.id.'],\
        ('ind_aln_id_ls', 0, ): ['', '', 1, 
            'a comma/dash-separated list of IndividualAlignment.id. '
            'This overrides ind_seq_id_ls.', ],\
        ('alignment_outdated_index', 0, int): [0, '', 1, 
            'filter based on value of IndividualAlignment.outdated_index.', ],\
        ("alignment_method_id", 0, int): [None, 'G', 1,
            'To filter alignments. None: whatever; integer: AlignmentMethod.id'],\
        ("local_realigned", 0, int): [None, '', 1, 
            'To filter which input alignments to fetch from db (i.e. '
            'AlignmentReadBaseQualityRecalibration.py)'
            'OR to instruct whether local_realigned should be applied '
            '(i.e. ShortRead2Alignment.py)'],\
        ('defaultSampleAlignmentDepth', 1, int): [10, '', 1, 
            "when database doesn't have median_depth info for one alignment, "
            "use this number instead.", ],\
        ('individual_sequence_file_raw_id_type', 1, int): [1, '', 1, 
            "1: only all-library-fused libraries,\n"
            "2: only library-specific alignments,\n"
            "3: both all-library-fused and library-specific alignments", ],\
        }
    option_default_dict.update(commonAlignmentWorkflowOptionDict)
    partitionWorkflowOptionDict= {
        ("selectedRegionFname", 0, ): ["", 'R', 1, 
            'the file is in bed format, tab-delimited, chr start stop. '
            'used to restrict SAMtools/GATK to only make calls at this region. '
            'start and stop are 0-based. i.e. start=0, stop=100 means bases from 0-99. '
            'This overrides the contig/chromosome selection approach defined'
            ' by --contigMaxRankBySize and --contigMinRankBySize. '
            'This file would be split into maxNoOfRegionsPerJob lines.'],\
        ('maxNoOfRegionsPerJob', 1, int): [5000, 'K', 1, 
            'Given selectedRegionFname, this dictates the maximum number of '
            'regions each job would handle. '
            'The actual number could be lower because the regions are first '
            'grouped into chromosomes. '
            'If one chromosome has <maxNoOfRegionsPerJob, then that job handles less.', ],\
        }
    option_default_dict.update(partitionWorkflowOptionDict)

    def __init__(self,
        pegasusFolderName='input',
        output_path=None,

        drivername='postgresql', hostname='localhost',
        dbname='', schema='public', port=None,
        db_user=None,
        db_passwd=None,
        data_dir=None, local_data_dir=None,

        ref_ind_seq_id=None,

        ind_seq_id_ls=None,
        ind_aln_id_ls=None,
        local_realigned=0,
        completedAlignment=None,
        skipDoneAlignment=False,
        excludeContaminant=False,
        sequence_filtered=None,

        alignment_outdated_index=0,
        alignment_method_id=None,
        defaultSampleAlignmentDepth=10,
        individual_sequence_file_raw_id_type=1,

        selectedRegionFname=None,
        maxNoOfRegionsPerJob=5000,
        
        samtools_path="bin/samtools",
        picard_dir="script/picard/dist",
        gatk_path="bin/GenomeAnalysisTK1_6_9.jar",
        gatk2_path="bin/GenomeAnalysisTK.jar",
        picard_path="script/picard.broad/build/libs/picard.jar",
        tabixPath="bin/tabix",
        vcftoolsPath="bin/vcftools/vcftools",
        ligateVcfPerlPath="bin/ligateVcf.pl",
        maxContigID=None,
        minContigID=None,
        contigMaxRankBySize=2500,
        contigMinRankBySize=1,

        chromosome_type_id=None, 
        ref_genome_tax_id=9606,
        ref_genome_sequence_type_id=1,
        ref_genome_version=15,
        ref_genome_outdated_index=0,

        mask_genotype_method_id=None, 
        checkEmptyVCFByReading=False,

        needFastaIndexJob=False,
        needFastaDictJob=False,
        reduce_reads=None,

        site_id_ls="",
        country_id_ls="",
        tax_id_ls="9606",
        sequence_type_id_ls="",
        sequencer_id_ls="",
        sequence_batch_id_ls="",
        version_ls="",
        sequence_min_coverage=None,
        sequence_max_coverage=None,
        
        alignmentDepthIntervalMethodShortName=None,
        minAlignmentDepthIntervalLength=1000,
        alignmentDepthMaxFold=2,
        alignmentDepthMinFold=0.1,

        intervalOverlapSize=500000,
        intervalSize=5000000,
        
        defaultGATKArguments=\
        " --unsafe ALL --validation_strictness SILENT --read_filter BadCigar ",
        
        tmpDir='/tmp/',
        max_walltime=4320,
        
        home_path=None,
        javaPath=None,
        pymodulePath="src/pymodule",
        thisModulePath=None,
        jvmVirtualByPhysicalMemoryRatio=1.2,
        
        site_handler='condor',
        input_site_handler='condor',
        cluster_size=30,

        needSSHDBTunnel=False,
        commit=False,
        debug=False, report=False):
        """
        """
        ParentClass.__init__(self,
            pegasusFolderName=pegasusFolderName,
            output_path=output_path,
            
            drivername=drivername, hostname=hostname,
            dbname=dbname, schema=schema, port=port,
            db_user=db_user, db_passwd=db_passwd,
            data_dir=data_dir, local_data_dir=local_data_dir,

            ref_ind_seq_id=ref_ind_seq_id,

            completedAlignment=completedAlignment,
            skipDoneAlignment=skipDoneAlignment,
            excludeContaminant=excludeContaminant,
            sequence_filtered=sequence_filtered,

            samtools_path=samtools_path,
            picard_dir=picard_dir,
            gatk_path=gatk_path,
            gatk2_path=gatk2_path,
            picard_path=picard_path,
            tabixPath=tabixPath,
            vcftoolsPath=vcftoolsPath,
            ligateVcfPerlPath=ligateVcfPerlPath,

            maxContigID=maxContigID,
            minContigID=minContigID,
            contigMaxRankBySize=contigMaxRankBySize,
            contigMinRankBySize=contigMinRankBySize,

            chromosome_type_id=chromosome_type_id, 
            ref_genome_tax_id=ref_genome_tax_id,
            ref_genome_sequence_type_id=ref_genome_sequence_type_id,
            ref_genome_version=ref_genome_version,
            ref_genome_outdated_index=ref_genome_outdated_index,
            
            mask_genotype_method_id=mask_genotype_method_id, 
            checkEmptyVCFByReading=checkEmptyVCFByReading,

            needFastaIndexJob=needFastaIndexJob,
            needFastaDictJob=needFastaDictJob,
            reduce_reads=reduce_reads,

            site_id_ls=site_id_ls,
            country_id_ls=country_id_ls,
            tax_id_ls=tax_id_ls,
            sequence_type_id_ls=sequence_type_id_ls,
            sequencer_id_ls=sequencer_id_ls,
            sequence_batch_id_ls=sequence_batch_id_ls,
            version_ls=version_ls,

            sequence_min_coverage=sequence_min_coverage,
            sequence_max_coverage=sequence_max_coverage,
            alignmentDepthIntervalMethodShortName=alignmentDepthIntervalMethodShortName,
            minAlignmentDepthIntervalLength=minAlignmentDepthIntervalLength,
            
            alignmentDepthMaxFold=alignmentDepthMaxFold,
            alignmentDepthMinFold=alignmentDepthMinFold,

            intervalOverlapSize=intervalOverlapSize,
            intervalSize=intervalSize,
            
            defaultGATKArguments=defaultGATKArguments,

            tmpDir=tmpDir,
            max_walltime=max_walltime, 
            home_path=home_path,
            javaPath=javaPath,
            pymodulePath=pymodulePath,
            thisModulePath=thisModulePath,
            jvmVirtualByPhysicalMemoryRatio=jvmVirtualByPhysicalMemoryRatio,
            
            site_handler=site_handler,
            input_site_handler=input_site_handler,
            cluster_size=cluster_size,

            needSSHDBTunnel=needSSHDBTunnel,
            commit=commit,
            debug=debug, report=report)
        
        self.ind_aln_id_ls = ind_aln_id_ls
        self.ind_seq_id_ls = ind_seq_id_ls
        self.local_realigned = local_realigned
        self.alignment_outdated_index = alignment_outdated_index
        self.alignment_method_id = alignment_method_id
        self.defaultSampleAlignmentDepth = defaultSampleAlignmentDepth
        self.individual_sequence_file_raw_id_type = individual_sequence_file_raw_id_type

        self.selectedRegionFname = selectedRegionFname
        self.maxNoOfRegionsPerJob = maxNoOfRegionsPerJob

        listArgumentName_data_type_ls = [('ind_seq_id_ls', int), 
            ("ind_aln_id_ls", int)]
        ProcessOptions.processListArguments(listArgumentName_data_type_ls, 
            emptyContent=[], class_to_have_attr=self)

        self.needSplitChrIntervalData = True
        self.mapReduceType = 1
        # type 1: split VCF with fixed number of sites
        # type 2: SelectVariants from VCF with fixed-size windows        
        # child classes can turn it off
        self.needGzipPreReduceReturnData = True
        self.needGzipReduceReturnData = True

    def addAlignmentAsInputToJobLs(self, alignmentDataLs=None, jobLs:List[Job]=None,
        jobInputOption=""):
        """
        Used in addGenotypeCallJobs() to add alignment files as input
            to calling jobs.
        """
        for alignmentData in alignmentDataLs:
            alignment = alignmentData.alignment
            bamF = alignmentData.bamF
            baiF = alignmentData.baiF
            for job in jobLs:
                if jobInputOption:
                    job.add_args(jobInputOption)
                job.add_args(bamF)
                #it's either symlink or stage-in
                self.addJobUse(job, bamF, is_input=True)
                self.addJobUse(job, baiF, is_input=True)
                self.add_dependency(job, parents=alignmentData.jobLs)

    def addAlignmentAsInputToPlatypusJobLs(self, alignmentDataLs=None,
        jobLs:List[Job]=None,
        jobInputOption="--bamFiles"):
        """
        2013.05.21 bugfix: pegasus/condor would truncate long single-argument.
        2013.05.16 different from addAlignmentAsInputToJobLs, that platypus is like this:
                --bamFiles=1.bam,2.bam,3.bam
            used in addGenotypeCallJobs() to add alignment files as input to calling jobs
        """
        for job in jobLs:
            if jobInputOption:
                job.add_args(jobInputOption)
            fileArgumentLs = []
            alignmentFileFolder = None
            for alignmentData in alignmentDataLs:
                alignment = alignmentData.alignment
                bamF = alignmentData.bamF
                baiF = alignmentData.baiF
                fileArgumentLs.append(bamF.name)
                if alignmentFileFolder is None:
                    alignmentFileFolder = os.path.split(bamF.name)[0]
                #it's either symlink or stage-in
                self.addJobUse(job, bamF, is_input=True)
                self.addJobUse(job, baiF, is_input=True)
                self.add_dependency(job, parents=alignmentData.jobLs)
            #if alignmentFileFolder:
            # #2013.05.21 pegasus/condor would truncate long single-argument.
            #	job.add_args('%s/*.bam'%(alignmentFileFolder))
            #else:
            job.add_args(','.join(fileArgumentLs))


    def addAddRG2BamJobsAsNeeded(self, alignmentDataLs=None, transferOutput=True):
        """
        2011-9-15
            add a read group only when the alignment doesn't have it according to db record
            DBVervet.pokeBamReadGroupPresence() from misc.py helps to fill in db records if it's unclear.
        2011-9-14
            The read-group adding jobs will have a "move" part that overwrites
                the original bam&bai if site_handler and input_site_handler is same.
            For those alignment files that don't need to. It doesn't matter.
             pegasus will transfer/symlink them.
        """
        print(f"Adding add-read-group2BAM jobs for {len(alignmentDataLs)} "
            f"alignments if read group is not detected ... ", flush=True)
        job_max_memory = 3500	#in MB

        addRG2BamDir = None
        addRG2BamDirJob = None

        no_of_rg_jobs = 0
        returnData = []
        for alignmentData in alignmentDataLs:
            alignment = alignmentData.alignment
            parentJobLs = alignmentData.jobLs
            bamF = alignmentData.bamF
            baiF = alignmentData.baiF
            if alignment.read_group_added!=1:
                if addRG2BamDir is None:
                    addRG2BamDir = "addRG2Bam"
                    addRG2BamDirJob = self.addMkDirJob(outputDir=addRG2BamDir)
                # add RG to this bam
                sequencer = alignment.individual_sequence.sequencer
                read_group = alignment.getReadGroup()
                if sequencer=='454':
                    platform_id = 'LS454'
                elif sequencer=='GA':
                    platform_id = 'ILLUMINA'
                else:
                    platform_id = 'ILLUMINA'
                outputRGSAM = File(os.path.join(addRG2BamDir,\
                    os.path.basename(alignment.path)))
                addRGJob:Job = self.addJavaJob(self.AddOrReplaceReadGroupsJava,
                    jarFile=self.AddOrReplaceReadGroupsJar,
                    inputArgumentOption="INPUT=", inputFile=bamF,
                    outputArgumentOption="OUTPUT=", outputFile=outputRGSAM,
                    transferOutput=transferOutput,
                    extraArgumentList=['RGID=%s'%(read_group), 
                        'RGLB=%s'%(platform_id),
                        'RGPL=%s'%(platform_id), 'RGPU=%s'%(read_group), 'RGSM=%s'%(read_group),
                        'SORT_ORDER=coordinate', "VALIDATION_STRINGENCY=LENIENT"],
                    parentJobLs=parentJobLs,
                    extraDependentInputLs=[baiF], job_max_memory=job_max_memory)
                    #(adding the SORT_ORDER doesn't do sorting but it marks the header
                    #  as sorted so that BuildBamIndexJar won't fail.)
                if self.tmpDir:
                    addRGJob.add_args("TMP_DIR=%s"%self.tmpDir)
                if addRG2BamDirJob:
                    self.add_dependency(addRGJob, parents=[addRG2BamDirJob])

                index_sam_job = self.addBAMIndexJob(
                    inputBamF=outputRGSAM, parentJobLs=[addRGJob],
                    transferOutput=transferOutput, javaMaxMemory=2000)
                newAlignmentData = PassingData(alignment=alignment)
                newAlignmentData.jobLs = [index_sam_job, addRGJob]
                newAlignmentData.bamF = index_sam_job.bamFile
                newAlignmentData.baiF = index_sam_job.baiFile
                no_of_rg_jobs += 1
            else:
                newAlignmentData = alignmentData
            returnData.append(newAlignmentData)
        print(f"{no_of_rg_jobs} alignments need read-group addition.",
            flush=True)
        return returnData

    def preReduce(self, passingData=None, transferOutput=True, **keywords):
        """
        setup additional mkdir folder jobs, before mapEachAlignment,
            mapEachChromosome, mapReduceOneAlignment
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        return returnData

    def mapEachChromosome(self, alignmentData=None, chromosome=None,\
        VCFJobData=None, passingData=None,
        reduceBeforeEachAlignmentData=None, transferOutput=True, **keywords):
        """
        2012.9.17
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        return returnData

    def map(self, alignmentData=None, intervalData=None,\
        VCFJobData=None, passingData=None,
        mapEachChromosomeData=None, transferOutput=True, **keywords):
        """
        2012.9.17
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        return returnData

    def mapEachInterval(self, **keywords):
        """
        2012.9.22 link to map()
        """
        return self.map(**keywords)


    def linkMapToReduce(self, mapEachIntervalData=None,
        preReduceReturnData=None, passingData=None, transferOutput=True, **keywords):
        """
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        return returnData

    def mapEachAlignment(self, alignmentData=None,  passingData=None,
        transferOutput=True, **keywords):
        """
        2012.9.22
            similar to reduceBeforeEachAlignmentData() but
             for mapping programs that run on one alignment each.

            passingData.alignmentJobAndOutputLs = []
            passingData.bamFnamePrefix = bamFnamePrefix
            passingData.individual_alignment = alignment
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []

        topOutputDirJob = passingData.topOutputDirJob
        refFastaF = passingData.refFastaFList[0]

        alignment = alignmentData.alignment
        parentJobLs = alignmentData.jobLs
        bamF = alignmentData.bamF
        baiF = alignmentData.baiF

        bamFnamePrefix = alignment.getReadGroup()

        return returnData

    def reduceAfterEachChromosome(self, chromosome=None, passingData=None,
        transferOutput=True,
        mapEachIntervalDataLs=None, **keywords):
        """
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        returnData.mapEachIntervalDataLs = mapEachIntervalDataLs
        return returnData

    def reduceBeforeEachAlignment(self, passingData=None,
        transferOutput=True, **keywords):
        """
        2012.9 setup some reduce jobs before loop over all intervals of one alignment begins.
            these reduce jobs will collect stuff from each map() job.
            the link will be established in linkMapToReduce().
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        return returnData

    def reduceAfterEachAlignment(self, passingData=None,
        mapEachChromosomeDataLs=None,
        reduceAfterEachChromosomeDataLs=None,\
        transferOutput=True, **keywords):
        """
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        returnData.mapEachChromosomeDataLs = mapEachChromosomeDataLs
        returnData.reduceAfterEachChromosomeDataLs = reduceAfterEachChromosomeDataLs
        return returnData

    def reduce(self, passingData=None, reduceAfterEachAlignmentDataLs=None,
            transferOutput=True, **keywords):
        """
        2012.9.17
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        returnData.reduceAfterEachAlignmentDataLs = reduceAfterEachAlignmentDataLs
        return returnData

    def mapReduceOneAlignment(self, alignmentData=None, passingData=None,
        chrIDSet=None, chrSizeIDList=None, chr2IntervalDataLs=None,
        chr2VCFJobData=None,
        outputDirPrefix=None, transferOutput=False,
        skipChromosomeIfVCFMissing=False, **keywords):
        """
        2013.04.11 moved from AbstractAlignmentAndVCFWorkflow.py
        2013.04.08, added skipChromosomeIfVCFMissing
        2013.1.25
        """
        returnData = PassingData()
        mapEachChromosomeDataLs = passingData.mapEachChromosomeDataLs
        mapEachChromosomeDataLs = []
        reduceBeforeEachAlignmentData = passingData.reduceBeforeEachAlignmentData
        mapEachAlignmentData = passingData.mapEachAlignmentData
        preReduceReturnData = passingData.preReduceReturnData

        for chromosomeSize, chromosome in chrSizeIDList:
            if chr2IntervalDataLs:
                intervalDataLs = chr2IntervalDataLs.get(chromosome, None)
            else:
                intervalDataLs = None
            if chr2VCFJobData:
                VCFJobData = chr2VCFJobData.get(chromosome)
            else:
                VCFJobData = None
            if VCFJobData is None:
                if self.report:
                    logging.warn(f"No VCFJobData for chromosome {chromosome}.")
                if skipChromosomeIfVCFMissing:
                    continue
                VCFJobData = PassingData(job=None, jobLs=[],
                    vcfFile=None, tbi_F=None, file=None, fileLs=[])
                VCFFile = None
            else:
                VCFFile = VCFJobData.file
                if VCFFile is None:
                    if self.report:
                        logging.warn(f"No VCFFile for chromosome {chromosome}.")
                    if skipChromosomeIfVCFMissing:
                        continue
            passingData.chromosome = chromosome
            mapEachChromosomeData = self.mapEachChromosome(
                alignmentData=alignmentData, chromosome=chromosome, \
                VCFJobData=VCFJobData, passingData=passingData,
                reduceBeforeEachAlignmentData=reduceBeforeEachAlignmentData,\
                mapEachAlignmentData=mapEachAlignmentData,\
                transferOutput=False, **keywords)
            passingData.mapEachChromosomeData = mapEachChromosomeData
            mapEachChromosomeDataLs.append(mapEachChromosomeData)

            mapEachIntervalDataLs = passingData.mapEachIntervalDataLs
            mapEachIntervalDataLs = []

            if intervalDataLs:
                for intervalData in intervalDataLs:
                    if intervalData.file:
                        mpileupInterval = intervalData.interval
                        bcftoolsInterval = intervalData.file
                    else:
                        mpileupInterval = intervalData.interval
                        bcftoolsInterval = intervalData.interval
                    intervalFileBasenameSignature = intervalData.intervalFileBasenameSignature
                    overlapInterval = intervalData.overlapInterval
                    overlapFileBasenameSignature = intervalData.overlapIntervalFileBasenameSignature

                    mapEachIntervalData = self.mapEachInterval(
                        alignmentData=alignmentData,
                        intervalData=intervalData,
                        chromosome=chromosome,\
                        VCFJobData=VCFJobData, passingData=passingData,
                        reduceBeforeEachAlignmentData=reduceBeforeEachAlignmentData,
                        mapEachAlignmentData=mapEachAlignmentData,\
                        mapEachChromosomeData=mapEachChromosomeData,
                        transferOutput=False, **keywords)
                    passingData.mapEachIntervalData = mapEachIntervalData
                    mapEachIntervalDataLs.append(mapEachIntervalData)

                    linkMapToReduceData = self.linkMapToReduce(
                        mapEachIntervalData=mapEachIntervalData, \
                        preReduceReturnData=preReduceReturnData, \
                        reduceBeforeEachAlignmentData=reduceBeforeEachAlignmentData,
                        mapEachAlignmentData=mapEachAlignmentData,\
                        passingData=passingData, \
                        **keywords)

            reduceAfterEachChromosomeData = self.reduceAfterEachChromosome(
                chromosome=chromosome,
                passingData=passingData,
                mapEachIntervalDataLs=passingData.mapEachIntervalDataLs,
                transferOutput=False, data_dir=self.data_dir, \
                **keywords)
            passingData.reduceAfterEachChromosomeData = reduceAfterEachChromosomeData
            passingData.reduceAfterEachChromosomeDataLs.append(
                reduceAfterEachChromosomeData)

            gzipReduceAfterEachChromosomeData = self.addGzipSubWorkflow(
                inputData=reduceAfterEachChromosomeData,
                transferOutput=transferOutput,
                outputDirPrefix="%sreduceAfterEachChromosome"%(outputDirPrefix),
                topOutputDirJob=passingData.gzipReduceAfterEachChromosomeFolderJob,
                report=False)
            passingData.gzipReduceAfterEachChromosomeFolderJob = \
                gzipReduceAfterEachChromosomeData.topOutputDirJob
        return returnData

    def setup_chr(self):
        """
        Use self.chr2size to derive chrIDSet.
        Added chrSizeIDList in return
            set chr2VCFJobData to None.
        """
        chrIDSet = set(self.chr2size.keys())
        chrSizeIDList = [(chromosomeSize, chromosome) for chromosome, 
            chromosomeSize in self.chr2size.items()]
        chrSizeIDList.sort()
        chrSizeIDList.reverse()
        #from big to small
        return PassingData(chrIDSet=chrIDSet, chr2VCFJobData=None,
            chrSizeIDList=chrSizeIDList)

    def addAllJobs(self,
        alignmentDataLs=None, chr2IntervalDataLs=None,
        skipDoneAlignment=False,\
        registerReferenceData=None, \
        needFastaIndexJob=False, needFastaDictJob=False, \
        data_dir=None, \
        outputDirPrefix="", transferOutput=True, **keywords):
        """
        2012.7.26
        """
        prePreprocessData = self.setup_chr()
        chrIDSet = prePreprocessData.chrIDSet
        chrSizeIDList = prePreprocessData.chrSizeIDList
        chr2VCFJobData = prePreprocessData.chr2VCFJobData

        print(f"Adding jobs that work on {len(alignmentDataLs)} alignments "
            f"(& possibly VCFs) for {len(chrIDSet)} chromosomes/contigs ...",
            flush=True)
        refFastaFList = registerReferenceData.refFastaFList
        refFastaF = refFastaFList[0]

        topOutputDirJob = self.addMkDirJob(outputDir="%sMap"%(outputDirPrefix))
        self.mapDirJob = topOutputDirJob

        plotOutputDirJob = self.addMkDirJob(outputDir="%sPlot"%(outputDirPrefix))
        self.plotOutputDirJob = plotOutputDirJob

        reduceOutputDirJob = self.addMkDirJob(outputDir="%sReduce"%(outputDirPrefix))
        self.reduceOutputDirJob = reduceOutputDirJob

        if needFastaDictJob or registerReferenceData.needPicardFastaDictJob:
            fastaDictJob = self.addRefFastaDictJob(
                refFastaF=refFastaF)
            refFastaDictF = fastaDictJob.refFastaDictF
        else:
            fastaDictJob = None
            refFastaDictF = registerReferenceData.refPicardFastaDictF

        if needFastaIndexJob or registerReferenceData.needSAMtoolsFastaIndexJob:
            fastaIndexJob = self.addRefFastaFaiIndexJob(refFastaF=refFastaF)
            refFastaIndexF = fastaIndexJob.refFastaIndexF
        else:
            fastaIndexJob = None
            refFastaIndexF = registerReferenceData.refSAMtoolsFastaIndexF

        returnData = PassingData()
        returnData.jobDataLs = []

        #2012.9.22 alignmentJobAndOutputLs is a relic.
        #	but it's similar to mapEachIntervalDataLs but
        #   designed for addAlignmentMergeJob(),
        #	so alignmentJobAndOutputLs gets re-set for every alignment.
        # 	mapEachAlignmentDataLs is never reset.
        #	mapEachChromosomeDataLs is reset right after a new alignment is chosen.
        #	mapEachIntervalDataLs is reset right after each chromosome is chosen.
        #	all reduce dataLs never gets reset.
        passingData = PassingData(alignmentJobAndOutputLs=[], \
            alignmentDataLs = alignmentDataLs,\
            bamFnamePrefix=None, \

            outputDirPrefix=outputDirPrefix, \
            topOutputDirJob=topOutputDirJob,\
            plotOutputDirJob=plotOutputDirJob,\
            reduceOutputDirJob = reduceOutputDirJob,\

            refFastaFList=refFastaFList, \
            registerReferenceData= registerReferenceData,\
            refFastaF=refFastaFList[0],\

            fastaDictJob = fastaDictJob,\
            refFastaDictF = refFastaDictF,\
            fastaIndexJob = fastaIndexJob,\
            refFastaIndexF = refFastaIndexF,\

            chromosome=None,\
            chrIDSet=chrIDSet,\
            chrSizeIDList = chrSizeIDList,\
            chr2IntervalDataLs=chr2IntervalDataLs,\

            mapEachAlignmentData = None,\
            mapEachChromosomeData=None, \
            mapEachIntervalData=None,\
            reduceBeforeEachAlignmentData = None, \
            reduceAfterEachAlignmentData=None,\
            reduceAfterEachChromosomeData=None,\

            mapEachAlignmentDataLs = [],\
            mapEachChromosomeDataLs=[], \
            mapEachIntervalDataLs=[],\
            reduceBeforeEachAlignmentDataLs = [], \
            reduceAfterEachAlignmentDataLs=[],\
            reduceAfterEachChromosomeDataLs=[],\

            gzipReduceAfterEachChromosomeFolderJob=None,\
            gzipReduceBeforeEachAlignmentFolderJob = None,\
            gzipReduceAfterEachAlignmentFolderJob = None,\
            gzipPreReduceFolderJob = None,\
            gzipReduceFolderJob=None,\
            )
        preReduceReturnData = self.preReduce(passingData=passingData,
            transferOutput=False, **keywords)
        passingData.preReduceReturnData = preReduceReturnData
        no_of_alignments_worked_on= 0
        for alignmentData in passingData.alignmentDataLs:
            alignment = alignmentData.alignment
            parentJobLs = alignmentData.jobLs + [fastaDictJob, fastaIndexJob]
            bamF = alignmentData.bamF
            baiF = alignmentData.baiF

            bamFnamePrefix = alignment.getReadGroup()

            passingData.alignmentJobAndOutputLs = []
            passingData.bamFnamePrefix = bamFnamePrefix
            passingData.individual_alignment = alignment
            passingData.alignmentData = alignmentData

            if skipDoneAlignment and self.isThisAlignmentComplete(
                individual_alignment=alignment, data_dir=data_dir):
                continue
            no_of_alignments_worked_on += 1
            mapEachAlignmentData = self.mapEachAlignment(
                alignmentData=alignmentData, passingData=passingData, \
                transferOutput=False, \
                preReduceReturnData=preReduceReturnData, **keywords)
            passingData.mapEachAlignmentDataLs.append(mapEachAlignmentData)
            passingData.mapEachAlignmentData = mapEachAlignmentData

            reduceBeforeEachAlignmentData = self.reduceBeforeEachAlignment(
                passingData=passingData,
                preReduceReturnData=preReduceReturnData, transferOutput=False, \
                **keywords)
            passingData.reduceBeforeEachAlignmentData = reduceBeforeEachAlignmentData
            passingData.reduceBeforeEachAlignmentDataLs.append(reduceBeforeEachAlignmentData)


            mapReduceOneAlignmentReturnData = self.mapReduceOneAlignment(
                alignmentData=alignmentData, \
                passingData=passingData, \
                chrIDSet=chrIDSet, chrSizeIDList=chrSizeIDList, \
                chr2IntervalDataLs=chr2IntervalDataLs,
                chr2VCFJobData=chr2VCFJobData,
                outputDirPrefix=outputDirPrefix, transferOutput=transferOutput)

            reduceAfterEachAlignmentData = self.reduceAfterEachAlignment(\
                mapEachAlignmentData=mapEachAlignmentData,\
                mapEachChromosomeDataLs=passingData.mapEachChromosomeDataLs,\
                reduceAfterEachChromosomeDataLs=passingData.reduceAfterEachChromosomeDataLs,\
                passingData=passingData, \
                transferOutput=False, data_dir=data_dir, **keywords)
            passingData.reduceAfterEachAlignmentData = reduceAfterEachAlignmentData
            passingData.reduceAfterEachAlignmentDataLs.append(reduceAfterEachAlignmentData)

            gzipReduceBeforeEachAlignmentData = self.addGzipSubWorkflow(\
                inputData=reduceBeforeEachAlignmentData, transferOutput=transferOutput,\
                outputDirPrefix="%sReduceBeforeEachAlignment"%(outputDirPrefix), \
                topOutputDirJob=passingData.gzipReduceBeforeEachAlignmentFolderJob, report=False)
            passingData.gzipReduceBeforeEachAlignmentFolderJob = \
                gzipReduceBeforeEachAlignmentData.topOutputDirJob

            gzipReduceAfterEachAlignmentData = self.addGzipSubWorkflow(\
                inputData=reduceAfterEachAlignmentData, transferOutput=transferOutput,\
                outputDirPrefix="%sReduceAfterEachAlignment"%(outputDirPrefix), \
                topOutputDirJob=passingData.gzipReduceAfterEachAlignmentFolderJob, \
                report=False)
            passingData.gzipReduceAfterEachAlignmentFolderJob = \
                gzipReduceAfterEachAlignmentData.topOutputDirJob
        reduceReturnData = self.reduce(passingData=passingData, \
            mapEachAlignmentData=passingData.mapEachAlignmentData, \
            reduceAfterEachAlignmentDataLs=passingData.reduceAfterEachAlignmentDataLs,\
            **keywords)
        passingData.reduceReturnData = reduceReturnData


        #2012.9.18 gzip the final output
        newReturnData = self.addGzipSubWorkflow(inputData=preReduceReturnData,
            transferOutput=transferOutput,\
            outputDirPrefix="%sGzipPreReduce"%(outputDirPrefix), \
            topOutputDirJob=passingData.gzipPreReduceFolderJob, \
            report=False)
        passingData.gzipPreReduceFolderJob = newReturnData.topOutputDirJob
        newReturnData = self.addGzipSubWorkflow(inputData=reduceReturnData,
            transferOutput=transferOutput,\
            outputDirPrefix="%sGzipReduce"%(outputDirPrefix), \
            topOutputDirJob=passingData.gzipReduceFolderJob, \
            report=False)
        passingData.gzipReduceFolderJob = newReturnData.topOutputDirJob

        sys.stderr.write("%s alignments to be worked on. %s jobs.\n"%(
            no_of_alignments_worked_on, self.no_of_jobs))
        return returnData


    def registerExecutables(self):
        """
        """
        ParentClass.registerExecutables(self)
        self.registerOneExecutable(path=self.samtools_path,
            name='samtools', clusterSizeMultiplier=0.2)
        self.samtoolsExecutableFile = self.registerOneExecutableAsFile(
            path=self.samtools_path,
            site_handler=self.input_site_handler)

    def setup_run(self):
        """
        Wrap all standard pre-run() related functions into this function.
            setting up for run(), called by run().
        """
        ParentClass.setup_run(self)
        # ParentClass.setup_run() will call getReferenceSequence() to
        #  setup self.registerReferenceData.
        
        if self.needSplitChrIntervalData:
            #2013.06.21 defined in ParentClass.__init__()
            if self.alignmentDepthIntervalMethodShortName and self.db_main and \
                self.db_main.checkAlignmentDepthIntervalMethod(
                    short_name=self.alignmentDepthIntervalMethodShortName):
                #2013.09.01 fetch intervals from db
                #make sure it exists in db first
                chr2IntervalDataLs = self.getChr2IntervalDataLsFromDBAlignmentDepthInterval(
                    db=self.db_main,
                    intervalSize=self.intervalSize,
                    intervalOverlapSize=self.intervalOverlapSize,
                    alignmentDepthIntervalMethodShortName=self.alignmentDepthIntervalMethodShortName,
                    alignmentDepthMinFold=self.alignmentDepthMinFold,
                    alignmentDepthMaxFold=self.alignmentDepthMaxFold,
                    minAlignmentDepthIntervalLength=self.minAlignmentDepthIntervalLength,
                    minContigID=self.minContigID,
                    maxContigID=self.maxContigID)
            else:
                #split evenly using chromosome size
                chr2IntervalDataLs = self.getChr2IntervalDataLsBySplitChrSize(
                    chr2size=self.chr2size,
                    intervalSize=self.intervalSize, \
                    intervalOverlapSize=self.intervalOverlapSize)
            # 2012.8.2 if maxContigID/minContigID is not well defined.
            #  restrictContigDictionry won't do anything.
            chr2IntervalDataLs = self.restrictContigDictionry(
                dc=chr2IntervalDataLs,
                maxContigID=self.maxContigID, minContigID=self.minContigID)
        else:
            logging.warn(f"self.needSplitChrIntervalData="\
                f"{self.needSplitChrIntervalData}, set chr2IntervalDataLs=None.")
            chr2IntervalDataLs = None

        alignmentLs = self.getAlignments()
        alignmentDataLs = self.registerAlignmentAndItsIndexFile(
            alignmentLs=alignmentLs, data_dir=self.data_dir)
        self.alignmentLs = alignmentLs
        self.alignmentDataLs = alignmentDataLs
        self.chr2IntervalDataLs = chr2IntervalDataLs
        return self

    def run(self):
        """
        2013.1.25
        """

        pdata = self.setup_run()
        self.addAllJobs(alignmentDataLs=pdata.alignmentDataLs,
            chr2IntervalDataLs=pdata.chr2IntervalDataLs,
            skipDoneAlignment=self.skipDoneAlignment,
            registerReferenceData=pdata.registerReferenceData,
            needFastaIndexJob=self.needFastaIndexJob,
            needFastaDictJob=self.needFastaDictJob,
            data_dir=self.data_dir, no_of_gatk_threads = 1,
            transferOutput=True,
            outputDirPrefix=self.pegasusFolderName)

        self.end_run()

if __name__ == '__main__':
    from argparse import ArgumentParser
    ap = ArgumentParser()
    ap.add_argument("--completedAlignment", type=int, default=None,
        help='To filter incomplete alignments: '
            '--completedAlignment 0 . '
            '--completedAlignment 1 fetches only the complete alignments. '
            'Default (%(default)s) has no effect.')


#!/usr/bin/env python3
"""
2007-10-23 A MPI wrapper
"""

import sys, pickle


def mpi_synchronize(communicator):
    """
    05-19-05
        copied from MpiBiclustering.py
    """
    sys.stdout.flush()
    sys.stderr.flush()
    communicator.barrier()


def fetch_cluster_block(curs, message_size, report=0):
    # 10-20-05 commonly used by input_node()
    if report:
        sys.stderr.write("Fetching stuff...\n")
    curs.execute("fetch 50 from crs")
    rows = curs.fetchall()
    prediction_ls = []
    string_length = 0
    while rows:
        for row in rows:
            prediction_ls.append(list(row))
            #the length to control MPI message size
            string_length += len(repr(row))
        if string_length>=message_size:
            break
        curs.execute("fetch 50 from crs")
        rows = curs.fetchall()
    if report:
        sys.stderr.write("Fetching done.\n")
    return prediction_ls


def mpi_schedule_jobs(communicator, job_list, node_function,
        node_parameter_list, debug=0):
    """
    05-19-05
        a universal scheduling function, the elements in job_list
        maybe string, integer, or something else, It's 'repr'ed before send.
        WARNING: NO -1 in job_list.
        So node_function should handle it as 'repr'ed.
        
        node_function()
            input: (element of the job_list, node_parameter_list).
            output: returns a value('repr'ed)
    """
    node_returned_value_list = []
    node_rank = communicator.rank
    if node_rank == 0:
        sys.stderr.write("\tTotally, %d jobs to be scheduled.\n"%len(job_list))
        seed_utilized = set()
        for node in range(1, communicator.size):
            if len(job_list)==0:	#if #nodes > #jobs, tell those nodes to break their listening loop.
                stop_signal = "-1"
                communicator.send(stop_signal, node, 0)	#no more jobs, stop that node,
                if debug:
                    sys.stderr.write("node %s stopped.\n"%node)
            else:
                job = job_list.pop(0)	#the first item poped first.
                communicator.send(repr(job), node, 0)	#string format
                if debug:
                    sys.stderr.write("node %s schedule a job, %s to %s\n"%(node_rank, repr(job), node))
                seed_utilized.add(node)
        
        #listen
        received_value, source, tag = communicator.receiveString(None, None)
        while received_value:
            node_returned_value_list.append(received_value)
            if len(job_list) == 0:	#first check if there're still files left, otherwise pop(0) raises error.
                stop_signal = "-1"
                communicator.send(stop_signal, source, 0)	#no more jobs, stop that node,
                if debug:
                    sys.stderr.write("node %s stopped.\n"%source)
                seed_utilized.remove(source)
                if len(seed_utilized) == 0:	#all seed used have finished their jobs
                    break
            else:
                job = job_list.pop(0)
                communicator.send(repr(job), source, 0)	#string format,
                if debug:
                    sys.stderr.write("node %s get one more job, %s\n"%(source, repr(job)) )
            received_value, source, tag = communicator.receiveString(None, None)	#listen
    else:
        received_data, source, tag = communicator.receiveString(0, None)	#get data from node 0,
        while received_data:
            if received_data=="-1":	#stop signal
                if debug:
                    sys.stderr.write("node %s breaked.\n"%node_rank)
                break
            else:
                sys.stderr.write("node %s working on %s...\n"%(node_rank, received_data))
                node_return_value = node_function(received_data, node_parameter_list)
                sys.stderr.write("node %s work on %s finished.\n"%(node_rank, received_data))
                communicator.send(repr(node_return_value), 0, node_rank)
                
            received_data, source, tag = communicator.receiveString(0, None)	#get data from node 0
    
    return node_returned_value_list

class MPIwrapper(object):
    """
    2008-05-08
        a general wrapper for mpi programs.
        
        functions copied from annot.bin.codense.common. made into a class.
    """
    def __init__(self, communicator, debug=0, report=0):
        self.communicator = communicator
        self.debug = debug
        self.report = report
    
    """
    10-21-05
        below output_node(), fetch_cluster_block(),input_node(), computing_node() are shared
        functions for Mpi Programs
    """
    def synchronize(self):
        """
        05-19-05
            copied from MpiBiclustering.py
        """
        sys.stdout.flush()
        sys.stderr.flush()
        self.communicator.barrier()
    
    def input_query_send_handler(self, param_obj, data_to_send):
        """
        2008-09-05
            to be called by input_node
            coming from variation/src/MpiQC149CrossMatch.py
            1. query the output_node to get a free computing node
            2. send the data to the computing node
        """
        #WATCH: tag is 1, to the output_node.
        self.communicator.send("1", param_obj.output_node_rank, 1)
        free_computing_node, source, tag = self.communicator.receiveString(param_obj.output_node_rank, 2)
        #WATCH: tag is 2, from the output_node
        data_pickle = pickle.dumps(data_to_send, -1)
        #WATCH: int()
        self.communicator.send(data_pickle, int(free_computing_node),0)
        if param_obj.report:
            sys.stderr.write("block %s sent to %s.\n"%(param_obj.counter, free_computing_node))
        param_obj.counter += 1
    
    def input_fetch_handler(self, param_obj, message_size=1, report=0):
        """
        2008-09-06
            coming from variation/src/MpiLD.py
            param_obj has 3 variables, params_ls, counter, report(optional)
            it pops a message_size number of entries out of param_obj.params_ls and then return it.
        """
        if param_obj.report:
            sys.stderr.write("Fetching stuff...\n")
        params_ls = param_obj.params_ls
        data_to_return = []
        for i in range(message_size):
            if len(params_ls)>0:
                one_parameter = params_ls.pop(0)
                data_to_return.append(one_parameter)
                param_obj.counter += 1
            else:
                break
        if getattr(param_obj, 'report', None):
            sys.stderr.write("Fetching done at counter=%s.\n"%(param_obj.counter))
        return data_to_return
    
    def output_node(self, free_computing_nodes, parameter_list, handler, type=1):
        """
        10-20-05
        10-21-05
            handle the situation when free_computing_node list is exhausted
            add type to specify the communicating data type
            based on type, different ways to get stop_signal
        10-21-05
            a common output_node() function
            two jobs:
            1. give node 0 the free_computing_node
            2. handle the data from the computing_nodes
            
            handler(communicator, parameter_list, data)
        """
        communicator = self.communicator
        report = self.report
        node_rank = communicator.rank
        sys.stderr.write("Node no.%s ready to accept output...\n"%node_rank)
        if type==1:
            data, source, tag = communicator.receiveString(None, 1)
        else:
            data, source, tag, count = communicator.receive(type, None, 1)
        no_of_computing_nodes = len(free_computing_nodes)
        no_of_resting_nodes = 0	#to keep track how many computing_nodes have rested
        free_computing_node_request = 0	#10-21-05 Flag used to check whether node 0 is requesting
        request_node = 0	#default request_node is 0
        while 1:
            if source==0:	#10-19-05 the input_node is asking me for free computing_node WATCH: it's array
                if len(free_computing_nodes)==0:	#10-21-05, it's empty. Flag the request.
                    free_computing_node_request = 1
                    request_node = source	#record the request_node
                else:
                    free_computing_node = free_computing_nodes.pop(0)
                    communicator.send(str(free_computing_node), source, 2)	#WATCH tag is 2.
            elif type==1 and data=="-1":
                no_of_resting_nodes += 1
                if report:
                    sys.stderr.write("node %s(%s-th) rested.\n"%(source, no_of_resting_nodes))
                if no_of_resting_nodes==no_of_computing_nodes:	#WATCH: its' size-3
                    break
                    if report:
                        sys.stderr.write("node %s output finished.\n"%node_rank)
            elif type!=1 and data.toscalar()==-1:	#10-21-05 for non-string type
                no_of_resting_nodes += 1
                if report:
                    sys.stderr.write("node %s(%s-th) rested.\n"%(source, no_of_resting_nodes))
                if no_of_resting_nodes==no_of_computing_nodes:	#WATCH: its' size-3
                    break
                    if report:
                        sys.stderr.write("node %s output finished.\n"%node_rank)
            else:
                if free_computing_node_request==1:	#there's a request for free_computing_node.
                    communicator.send(str(source), request_node, 2)	#WATCH tag is 2.
                    free_computing_node_request = 0
                else:
                    free_computing_nodes.append(source)	#append the free computing_node
                handler(communicator, parameter_list, data)
            if type==1:
                data, source, tag = communicator.receiveString(None, 1)
            else:
                data, source, tag, count = communicator.receive(type, None, 1)
        sys.stderr.write("Node no.%s output done.\n"%node_rank)
    
    def input_node(self, curs, free_computing_nodes, message_size=None, input_handler=fetch_cluster_block):
        """
        10-20-05
        10-22-05
            add input_handler and regard curs as parameter_list
        """
        communicator = self.communicator
        node_rank = communicator.rank
        sys.stderr.write("Input node(%s) working...\n"%node_rank)
        data = input_handler(curs, message_size, self.report)
        counter = 0
        while data:
            communicator.send("1", communicator.size-1, 1)	#WATCH: tag is 1, to the output_node.
            free_computing_node, source, tag = communicator.receiveString(communicator.size-1, 2)
                #WATCH: tag is 2, from the output_node
            data_pickle = pickle.dumps(data, -1)
            communicator.send(data_pickle, int(free_computing_node), 0)	#WATCH: int()
            if self.report:
                sys.stderr.write("block %s sent to %s.\n"%(counter, free_computing_node))
            data = input_handler(curs, message_size, self.report)
            counter += 1
        #tell computing_node to exit the loop
        for node in free_computing_nodes:	#send it to the computing_node
            if self.report:
                sys.stderr.write("exit signal sent to %s.\n"%(node))
            communicator.send("-1", node, 0)
        sys.stderr.write("Input node(%s) done\n"%(node_rank))
    
    def inputNode(self, param_obj, free_computing_nodes, param_generator=None, message_size=1):
        """
        2010-8-8
            Stop assuming that param_generator has method 'next'.
            Probe its attributes first. Use 'pop' if it has 'pop' and not 'next'.
        2010-5-31
            set message_size default to 1
            add argument message_size to allow sending >1 param to computing node
        2009-2-10
            version similar to input_node() but use a generator (argument param_generator)
        """
        communicator = self.communicator
        node_rank = communicator.rank
        sys.stderr.write("Input node(%s) working...\n"%node_rank)
        counter = 0
        timeToBreak = 0
        generatorMethod = 'next'
        if not hasattr(param_generator, generatorMethod) and hasattr(param_generator, 'pop'):
            generatorMethod = 'pop'
        while 1:
            param_ls = []
            for i in range(message_size):
                try:
                    param = getattr(param_generator, generatorMethod)()
                    param_ls.append(param)
                except:
                    if self.report:
                        sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
                        import traceback
                        traceback.print_exc()
                        sys.stderr.write('Input node broke out due to except.\n')
                    timeToBreak = 1
                    break
            if len(param_ls)>0:
                communicator.send("1", param_obj.output_node_rank, 1)	#WATCH: tag is 1, to the output_node.
                free_computing_node, source, tag = communicator.receiveString(param_obj.output_node_rank, 2)
                    #WATCH: tag is 2, from the output_node
                data_pickle = pickle.dumps(param_ls, -1)
                communicator.send(data_pickle, int(free_computing_node), 0)	#WATCH: int()
                if self.report:
                    sys.stderr.write("block %s sent to %s.\n"%(counter, free_computing_node))
                counter += 1
            if timeToBreak:
                break
            
        #tell computing_node to exit the loop
        for node in free_computing_nodes:	#send it to the computing_node
            if self.report:
                sys.stderr.write("exit signal sent to %s.\n"%(node))
            communicator.send("-1", node, 0)
        sys.stderr.write("Input node(%s) done\n"%(node_rank))
    
    
    def computing_cleanup_handler(self, communicator, output_node_rank=None):
        """
        2010-5-31
            not a classmethod anymore. add argument output_node_rank
        10-22-05
            default cleanup_handler
        """
        if self.report:
            sys.stderr.write("node %s telling %s that it's quitting.\n"%(communicator.rank, output_node_rank))
        #tell the last node to stop
        communicator.send("-1", output_node_rank, 1)	#tag is 1
    
    def computing_node(self, parameter_list, node_fire_handler, cleanup_handler=None, output_node_rank=None, inputNodeRank=None):
        """
        2010-5-30
            add more reporting if self.report is on. add argument output_node_rank & inputNodeRank.
        2008-09-08
            delete result and result_pickle right after they are useless to free up memory
        10-21-05
            0 is the node where the data is from
            size-1 is the node where output goes
        10-22-05
            make computing_cleanup_handler to be the default cleanup_handler
        """
        report = self.report
        communicator = self.communicator
        node_rank = communicator.rank
        if inputNodeRank is None:
            inputNodeRank = 0
        data, source, tag = communicator.receiveString(inputNodeRank, 0)	#get data from node 0
        if output_node_rank is None:
            output_node_rank = communicator.size-1
        while 1:
            if data=="-1":
                if self.report:
                    sys.stderr.write("node %s broke.\n"%node_rank)
                break
            else:
                result = node_fire_handler(communicator, data, parameter_list)
                if self.report:
                    sys.stderr.write("node %s is pickling result ...\n"%node_rank)
                result_pickle = pickle.dumps(result, -1)
                del result
                if self.report:
                    sys.stderr.write("node %s result-pickling done and sending it to node %s.\n"%(node_rank, output_node_rank))
                communicator.send(result_pickle, output_node_rank, 1)
                del result_pickle
                if self.report:
                    sys.stderr.write("node %s result-sending done.\n"%(node_rank))
            data, source, tag = communicator.receiveString(inputNodeRank, 0)	#get data from node 0
        if not cleanup_handler:
            self.computing_cleanup_handler(communicator, output_node_rank=output_node_rank)

class MPI4pywrapper(MPIwrapper):
    """
    2011
        a derivative of MPIwrapper, for mpi4py
            
            1. no more pickling
            2. recv() replaces receiveString()
            
                syntax difference
            
                to receive something from anywhere in mpi4py:
                    status = MPI.Status()
                    any_src, any_tag = MPI.ANY_SOURCE, MPI.ANY_TAG
                    data = communicator.recv(source=any_src, tag=any_tag, status=status)
                    source = status.source
        
    """
    def __init__(self, communicator, debug=0, report=0):
        MPIwrapper.__init__(self, communicator, debug=debug, report=report)
    
    def inputNode(self, param_obj, free_computing_nodes, param_generator=None, message_size=1):
        """
        2011-2-10
            derived from MPIwrapper.inputNode
            
            1. no more pickling
            2. recv() replaces receiveString()
        """
        communicator = self.communicator
        node_rank = communicator.rank
        sys.stderr.write("Input node(%s) working...\n"%node_rank)
        counter = 0
        timeToBreak = 0
        generatorMethod = 'next'
        if not hasattr(param_generator, generatorMethod) and hasattr(param_generator, 'pop'):
            generatorMethod = 'pop'
        while 1:
            param_ls = []
            for i in range(message_size):
                try:
                    param = getattr(param_generator, generatorMethod)()
                    param_ls.append(param)
                except:
                    if self.report:
                        sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
                        import traceback
                        traceback.print_exc()
                        sys.stderr.write('Input node broke out due to except.\n')
                    timeToBreak = 1
                    break
            if len(param_ls)>0:
                if self.report:
                    sys.stderr.write('Requesting free node from %s \n'%param_obj.output_node_rank)
                communicator.send("1", param_obj.output_node_rank, 1)	#WATCH: tag is 1, to the output_node.
                free_computing_node = communicator.recv(source=param_obj.output_node_rank, tag=2)
                if self.report:
                    sys.stderr.write('node %s is free \n'%free_computing_node)
                    #WATCH: tag is 2, from the output_node
                #data_pickle = pickle.dumps(param_ls, -1)
                communicator.send(param_ls, int(free_computing_node), 0)	#WATCH: int()
                if self.report:
                    sys.stderr.write("block %s sent to %s.\n"%(counter, free_computing_node))
                counter += 1
            if timeToBreak:
                break
            
        #tell computing_node to exit the loop
        for node in free_computing_nodes:	#send it to the computing_node
            if self.report:
                sys.stderr.write("exit signal sent to %s.\n"%(node))
            communicator.send("-1", node, 0)
        sys.stderr.write("Input node(%s) done\n"%(node_rank))
    
    
    def computing_node(self, parameter_list, node_fire_handler, cleanup_handler=None, output_node_rank=None, inputNodeRank=None):
        """
        2011-2-10
            1. no more pickling
            2. recv() replaces receiveString()
        """
        report = self.report
        communicator = self.communicator
        node_rank = communicator.rank
        if inputNodeRank is None:
            inputNodeRank = 0
        data = communicator.recv(source=inputNodeRank, tag=0)	#get data from node 0
        if output_node_rank is None:
            output_node_rank = communicator.size-1
        while 1:
            if data=="-1":
                if self.report:
                    sys.stderr.write("node %s broke.\n"%node_rank)
                break
            else:
                result = node_fire_handler(communicator, data, parameter_list)
                if self.report:
                    sys.stderr.write("node %s is pickling result ...\n"%node_rank)
                #result_pickle = pickle.dumps(result, -1)
                #del result
                if self.report:
                    sys.stderr.write("node %s result-pickling done and sending it to node %s.\n"%(node_rank, output_node_rank))
                communicator.send(result, output_node_rank, 1)
                #del result_pickle
                if self.report:
                    sys.stderr.write("node %s result-sending done.\n"%(node_rank))
            data = communicator.recv(source=inputNodeRank, tag=0)	#get data from node 0
        if not cleanup_handler:
            self.computing_cleanup_handler(communicator, output_node_rank=output_node_rank)
    
    def output_node(self, free_computing_nodes, parameter_list, handler, type=1):
        """
        2011-2-10
            1. no more pickling
            2. recv() replaces receiveString()
        """
        communicator = self.communicator
        report = self.report
        node_rank = communicator.rank
        from mpi4py import MPI		
        status = MPI.Status()
        any_src, any_tag = MPI.ANY_SOURCE, MPI.ANY_TAG
        sys.stderr.write("Node no.%s ready to accept output...\n"%node_rank)
        data = communicator.recv(source=any_src, tag=any_tag, status=status)
        source = status.source
        no_of_computing_nodes = len(free_computing_nodes)
        no_of_resting_nodes = 0	#to keep track how many computing_nodes have rested
        free_computing_node_request = 0	#10-21-05 Flag used to check whether node 0 is requesting
        request_node = 0	#default request_node is 0
        while 1:
            if source==0:	#10-19-05 the input_node is asking me for free computing_node WATCH: it's array
                if self.report:
                    sys.stderr.write("node %s request a free computing node.\n"%(source,))
                if len(free_computing_nodes)==0:	#10-21-05, it's empty. Flag the request.
                    if self.report:
                        sys.stderr.write("no free computing node available now.\n")
                    free_computing_node_request = 1
                    request_node = source	#record the request_node
                else:
                    free_computing_node = free_computing_nodes.pop(0)
                    if self.report:
                        sys.stderr.write("free computing node %s is  available.\n"%(free_computing_node))
                    communicator.send(str(free_computing_node), dest=source, tag=2)	#WATCH tag is 2.
            elif data=="-1":
                no_of_resting_nodes += 1
                if self.report:
                    sys.stderr.write("node %s(%s-th) rested.\n"%(source, no_of_resting_nodes))
                if no_of_resting_nodes==no_of_computing_nodes:	#WATCH: its' size-3
                    break
                    if self.report:
                        sys.stderr.write("node %s output finished.\n"%node_rank)
            else:
                if free_computing_node_request==1:	#there's an old request for free_computing_node.
                    if self.report:
                        sys.stderr.write("send the newly-free node %s to node %s.\n"%(source, request_node))
                    communicator.send(str(source), dest=request_node, tag=2)	#WATCH tag is 2.
                    free_computing_node_request = 0
                else:
                    if self.report:
                        sys.stderr.write("node %s has finished its job.\n"%(source,))
                    free_computing_nodes.append(source)	#append the free computing_node
                handler(communicator, parameter_list, data)
            data = communicator.recv(source=any_src, tag=any_tag, status=status)
            source = status.source
        sys.stderr.write("Node no.%s output done.\n"%node_rank)		



#!/usr/bin/env python3
"""
Description:
    2012.3.6 an abstract class for mapper
"""

import sys, os, math
import csv
from palos import ProcessOptions, getListOutOfStr, PassingData, utils
import re

class AbstractMapper(object):
    __doc__ = __doc__
    db_option_dict = {
        ('drivername', 1,):['postgresql', 'v', 1,	
            'which type of database? mysql or postgresql. For SQLAlchemy', ],
        ('hostname', 1, ): ['localhost', 'z', 1, 'hostname of the db server', ],
        ('dbname', 1, ): ['vervetdb', 'd', 1, 'database name', ],\
        ('schema', 0, ): [None, '', 1, 'database schema name', ],\
        ('db_user', 1, ): [None, 'u', 1, 'database username', ],\
        ('db_passwd', 1, ): [None, '', 1, 'database password', ],\
        ('port', 0, ):[None, '', 1, 'database port number'],\
        ('commit', 0, int):[0, '', 0, 'commit db transaction'],\
        ('data_dir', 0, ):[None, '', 1, 'path to the db-affliated file storage.'],
        }
    genome_db_option_dict = {
        ('genome_drivername', 1,):['postgresql', '', 1, 
            'which type of database: mysql or postgresql',],
        ('genome_hostname', 1, ): ['uclaOffice', '', 1,
            'hostname of the genome db server', ],
        ('genome_dbname', 1, ): ['vervetdb', '', 1, 'genome database name', ],
        ('genome_schema', 0, ): ['genome', '', 1, 
            'genome database schema name',],
        ('genome_db_user', 1, ): ['yh', '', 1, 'genome database username', ],
        ('genome_db_passwd', 1, ): [None, '', 1, 'genome database password',],
        ('genome_port', 0, ):[None, '', 1, 'genome database port number'],
        }
    option_default_dict = {
        ('inputFname', 0, ): [None, 'i', 1, 'The input file.', ],\
        ('inputDelimiter', 0, ): [None, '', 1, 
            'Delimiter of input file: tab, space, or coma. '
            'If not given, it will guess.', ],\
        ("home_path", 1, ): [os.path.expanduser("~"), 'e', 1, \
            'path to the home directory on the working nodes'],
        ('outputFname', 0, ): [None, 'o', 1, 'output file'],\
        ('outputFnamePrefix', 0, ): [None, 'O', 1, \
            'output filename prefix (optional).'],
        ('debug', 0, int):[0, 'b', 0, 'toggle debug mode'],
        ('report', 0, int):[0, 'r', 0, 
            'toggle report, more verbose stdout/stderr.']
        }

    def __init__(self, inputFnameLs=None, **keywords):
        """
        """
        self.ad = ProcessOptions.process_function_arguments(keywords, \
            self.option_default_dict, error_doc=self.__doc__, \
            class_to_have_attr=self)
        if getattr(self, 'outputFname', None) and hasattr(self,'outputFnamePrefix'):
            self.outputFnamePrefix = os.path.splitext(self.outputFname)[0]
        
        self.inputFnameLs = inputFnameLs
        if self.inputFnameLs is None:
            self.inputFnameLs = []
        
        if getattr(self, 'inputFname', None):
            self.inputFnameLs.insert(0, self.inputFname)
        
        #2013.08.14 setup self.inputFname if self.inputFnameLs is not None
        if self.inputFnameLs and not getattr(self, 'inputFname', None):
            self.inputFname = self.inputFnameLs[0]
        
        if getattr(self, 'inputDelimiter', None):
            if self.inputDelimiter=='tab':
                self.inputDelimiter = '\t'
            if self.inputDelimiter=='space':
                self.inputDelimiter = ' '
            if self.inputDelimiter=='coma':
                self.inputDelimiter = ','
        self.connectDB()
        #2012.8.14 an expression that searches any character in a string
        self.p_char = re.compile(r'[a-df-zA-DF-Z\-]$')
        #no 'e' or 'E', used in scientific number, add '-' and append '$'
    
    def insertHomePath(self, inputPath, home_path):
        """
        2012.8.19 copied from AbstractWorkflow
        """
        if inputPath.find('%s')!=-1:
            inputPath = inputPath%home_path
        return inputPath
    
    def connectDB(self):
        """
        2012.5.11
            place holder.
        """
        pass
        """
        db_main = SunsetDB.SunsetDB(drivername=self.drivername, \
            hostname=self.hostname, database=self.dbname, schema=self.schema,
            port=self.port,
            username=self.db_user, password=self.db_passwd)
        db_main.setup(create_tables=False)
        self.db_main = db_main
        """
    
    def run(self):
        """
        2013.05.01 place holder
        """
        pass

if __name__ == '__main__':
    main_class = AbstractMapper
    po = ProcessOptions(sys.argv, main_class.option_default_dict,
        error_doc=main_class.__doc__)
    instance = main_class(**po.long_option2value)
    instance.run()	


# Roman numeral conversion
# Copyright (c) 2001 Mark Pilgrim
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GPL-compatible Python 1.6.1 license:
#	 http://hdl.handle.net/1895.22/1013

# We'll be using regular expressions to match partial Roman numerals
# in fromRoman.  Read all about regular expressions in Python here:
#	 http://py-howto.sourceforge.net/regex/regex.html
import re

# What we really want here is an ordered structure of related elements,
# so we're storing them in a tuple of tuples.  Tuples are faster than
# lists but can be iterated through in the same manner, and they're
# perfect for static data structures like this.  We do not want to
# use dictionaries for this, because dictionaries do not retain their
# defined order (mapping have no order).
# The re.compile stuff pre-compiles a regular expression that we'll
# use in fromRoman to detect any valid number of occurrences of each
# symbol as we scan through the Roman numeral.
rom = (('M', 1000, re.compile('^MM?M?')), 
	('CM', 900, re.compile('^CM')), 
	('D', 500, re.compile('^D')), 
	('CD', 400, re.compile('^CD')), 
	('C', 100, re.compile('^CC?C?')), 
	('XC', 90, re.compile('^XC')), 
	('L', 50, re.compile('^L')), 
	('XL', 40, re.compile('^XL')), 
	('X', 10, re.compile('^XX?X?')), 
	('IX', 9, re.compile('^IX')), 
	('V', 5, re.compile('^V')), 
	('IV', 4, re.compile('^IV')), 
	('I', 1, re.compile('^II?I?')))

def toRoman(n):
	result = ""
	
	# assert is a Python built-in that takes 2 arguments: an expression
	# to evaluate, and an error message.
	# If the expression evaluates to 0, '', [], (), {}, None, or any other
	# false value, assert will raise an AssertionError with the given
	# error message.
	assert 0 < n < 4000, "number out of range (must be 1..3999)"
	
	# We can quickly iterate through our data structure (a tuple of tuples)
	# all pull out each value from each tuple in turn.  Called
	# "tuple unpacking" or "multi-variable assignment".  See
	#	 http://diveintopython.org/odbchelper_multiassign.html
	for roman, arabic, pattern in rom:
		while n >= arabic:
			# Python 2.0 supports augmented assignment, and it's
			# faster than doing result = result + roman.
			result += roman
			n -= arabic
	return result
	
def fromRoman(s):
	result = 0
	# We'll be munging s as we scan through it, so save the original version
	# for error reporting later.
	original = s
	
	# Python 2.0 supports string methods.  In earlier versions, you would
	# do this instead:
	#	 import string
	#	 s = string.upper(s)
	s = s.upper()
	
	for roman, arabic, pattern in rom:
		# pattern is the pre-compiled regular expression pattern,
		# which has methods to search a given string for the
		# regular expression the pattern was compiled with.
		match = pattern.search(s)

		# If the regular expression engine found a match,
		# the search method returns a MatchObject, otherwise
		# it returns None.  None is a false value in Python,
		# but a MatchObject (and almost any other class instance)
		# is true, so we can just say "if match:" here instead of
		# "if match <> None:" or some other more verbose incarnation.
		if match:
			# The magic starts here.  If we found a match,
			# what we found was 1 or more occurrences of the
			# same Roman numeral digit ("M", or "CM", or "D", or
			# whatever) in our data structure.  That means that
			# the arabic value of this found fragment is the
			# arabic value of a single digit * the number of
			# digits found.  match.group() gives us the found
			# fragment, but the length of that string may not
			# be the number of digits found because individual
			# Roman digits can be more than 1 character (e.g. "CM").
			# So the total arabic value is the arabic value of the
			# digit, times the character length of the found fragment,
			# divided by the character length of the digit.
			lenmatch = len(match.group())
			result += arabic * lenmatch / len(roman)
			
			# Now hack the processed part off the beginning of
			# the string and continue.
			s = s[lenmatch:]
	
	# 1. If s has anything left, that means the Roman numeral was mal-formed, 
	# because otherwise our regular expressions would have caught it.  This
	# depends on the fact that Roman numerals have to use the highest digits
	# possible, and our data structure was ordered from highest to lowest.
	# 2. locals() is a function that returns a dictionary of local variables, 
	# with keys being the variable names as strings and values being the
	# actual variables' values.  See
	#	 http://diveintopython.org/dialect_locals.html
	# 3. The %(original)s is dictionary-based string formatting;
	# it gets replaced by the value of the 'original' key in the given
	# dictionary; since the given dictionary is a dictionary of local
	# variables, this effectively inserts the value of <original> into
	# the string.  See
	#	 http://diveintopython.org/dialect_dictsub.html
	assert s == '', "invalid Roman numeral: %(original)s" % locals()
	
	return result
	
def test():
	# If you're serious about unit testing, check out PyUnit, which will be
	# part of the standard library in Python 2.1:
	#	 http://pyunit.sourceforge.net/

	# Test all the good values.  range is a built-in Python function that
	# returns a list of integers starting with the first parameter and
	# up to but not including the second parameter.  This is akin to
	# for (n=1; n<4000; n++) { ... } in C.
	for n in range(1, 4000):
		s = toRoman(n)
		fullcircle = fromRoman(s)
		assert n == fullcircle, 'FAILED: %(n)d --> %(s)s --> %(fullcircle)s'%locals()
		
	# Test some bad values for the arabic --> Roman conversion.
	# We expect the function to raise an AssertionError, so we
	# explicitly catch that; if toRoman doesn't raise an exception,
	# we treat that as an error and raise our own expection in response.
	# For more on exceptions, see
	#	 http://diveintopython.org/fileinfo_exception.html
	# Note that we're calling a function which is returning a result,
	# but we're never assigning the result to anything.  That's OK;
	# Python will just swallow the result and continue without error.
	for n in (4000, 0, -1):
		try:
			toRoman(n)
		except AssertionError:
			pass
		else:
			raise AssertionError, "FAILED: toRoman(%(n)d) should have failed but didn't" % locals()
	
	# Test some bad values for the Roman --> arabic conversion.
	for s in ('MMMM', 'CMCM', 'DD', 'CDCD', 'CCCC', 'XCXC', 'LL', 'XLXL', 
'XXXX', 
			'IXIX', 'VV', 'IVIV', 'IIII', 'IIMXCC', 'CMM'):
		try:
			fromRoman(s)
		except AssertionError:
			pass
		else:
			raise AssertionError, "FAILED: fromRoman('%(s)s') should have failed but didn't" % locals()
	
	# If we got this far, every test passed (otherwise we would have
	# raised an exception by now, or one of the functions we called would
	# have raised one and we intentionally wouldn't have caught it,
	# so we'd never get here).
	print "Test suite completed successfully."
	
# This is a standard trick in Python.  If this script is imported by another
# module, this module's __main__ attribute will be some descriptive value
# (generally the filename minus the .py extension).  But if the script is
# run by itself from the command line (or through an IDE), the module's
# __name__ attribute will be a special value, "__main__".  We can use this
# discrepancy to our advantage and define a test suite that will only run
# when the script is run by itself; the test suite will be completely ignored
# when the module is imported by another module.  See
#	 http://diveintopython.org/odbchelper_testing.html
if __name__ == "__main__":
	test()
	roman_str = 'VI'
	print '%s is %s.'%(roman_str, fromRoman(roman_str))

#---end of roman.py



#!/usr/bin/env python3
"""
2008-11-18
A module to do PCA.
"""

import sys, os, math
import numpy

class PCA(object):
	__doc__ = __doc__
	option_default_dict = {('input_fname', 1,):['', 'i', 1, 'input file', ],\
							('commit', 0, int):[0, 'c', 0, 'commit the db operation. this commit happens after every db operation, not wait till the end.'],\
							('debug', 0, int):[0, 'b', 0, 'debug mode. 1=level 1 (pdb mode). 2=level 2 (same as 1 except no pdb mode)'],\
							('report', 0, int):[0, 'r', 0, 'toggle report, more verbose stdout/stderr.']}
	def __init__(self,  **keywords):
		"""
		2008-11-18
		"""
		from palos import ProcessOptions
		self.ad = ProcessOptions.process_function_arguments(keywords, self.option_default_dict, error_doc=self.__doc__, class_to_have_attr=self)
	
	def normalize(cls, data_matrix, divide_variance=True):
		"""
		2016.04.14 bugfix, computing using new_data_matrix instead of data_matrix
		2008-11-18
			this is based on Patterson2006a.
			
			assume no missing data in data_matrix. SNP allele in data_matrix is 0 or 1. two chromosomes are homozygous.
			so each strain has either allele 0 or 1 on both chromosomes. (selfing plants, like arabidopsis thaliana)
			
			normalizing steps:
				1. subtract mean from each column
				2. make variance equal among columns by dividing each column by the estimated stdev
					if variance is 0, skip this step.
		"""
		sys.stderr.write("Normalizing ...")
		no_of_rows, no_of_cols = data_matrix.shape
		new_data_matrix = data_matrix.astype(numpy.float)	#change the data type
		
		for j in range(no_of_cols):
			columnDataVector = new_data_matrix[:,j]
			col_mean = numpy.mean(columnDataVector)
			col_var = numpy.var(columnDataVector)
			#col_var = col_mean*(1-col_mean)	#2009-9-3 only good for binary matrix
			if col_mean!=0 and divide_variance and col_var!=0:
				new_data_matrix[:,j] = (new_data_matrix[:,j]-col_mean)/numpy.sqrt(col_var)
			else:
				new_data_matrix[:,j] = new_data_matrix[:,j]-col_mean
		sys.stderr.write("Done.\n")
		return new_data_matrix
	normalize = classmethod(normalize)
	
	def eig(cls, data_matrix, toNormalize=True):
		"""
		2008-11-18
			numpy.inner(matrix1, matrix2) is weird. matrix1's 2nd dimension is equal to matrix2's 2nd dimension.
				Not as traditional, matrix1's 2nd dimension is equal to matrix2's 1st dimension.
		"""
		if toNormalize:
			new_data_matrix = cls.normalize(data_matrix)
			new_data_matrix2 = numpy.transpose(new_data_matrix)	#2008-11-20 transpose only to get cov_matrix. a later mulitplication between eigen_vector and new_data_matrix
			cov_matrix = 1.0/new_data_matrix2.shape[1]*numpy.inner(new_data_matrix2, new_data_matrix2)	#2008-11-19 mysteriously, numpy.tranpose() is not required on the 2nd new_data_matrix, it'll cause ValueError('matrices are not aligned',)
		else:
			new_data_matrix = cls.normalize(data_matrix, divide_variance=False)
			#new_data_matrix = data_matrix
			cov_matrix = numpy.cov(new_data_matrix, rowvar=0)	#2009-9-3 numpy.cov or numpy.corrcoef
		#eigen_values, eigen_vectors = numpy.linalg.eig(cov_matrix)	#get complex values out of this
		import rpy
		eigen_result = rpy.r.eigen(cov_matrix)
		eigen_values = numpy.array(eigen_result['values'])
		eigen_vectors = eigen_result['vectors']
		explained_var = eigen_values/numpy.sum(eigen_values)
		pc_matrix = numpy.inner(numpy.transpose(eigen_vectors), new_data_matrix)	#eigen_vectors has to be transposed in row-vector form
		pc_matrix = numpy.transpose(pc_matrix)	#transpose again. maybe pc_matrix = numpy.inner(new_data_matrix, eigen_vectors)
		return pc_matrix, eigen_vectors, explained_var
	
	eig = classmethod(eig)

if __name__ == '__main__':
	from palos import ProcessOptions
	main_class = PCA
	#po = ProcessOptions(sys.argv, main_class.option_default_dict, error_doc=main_class.__doc__)
	
	#instance = main_class(**po.long_option2value)
	#instance.run()


#!/usr/bin/python
#
# Copyright (C) 2008 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Converts Python data into data for Google Visualization API clients.

This library can be used to create a google.visualization.DataTable usable by
visualizations built on the Google Visualization API. Output formats are raw
JSON, JSON response, and JavaScript.

See http://code.google.com/apis/visualization/ for documentation on the
Google Visualization API.
"""

__author__ = "Amit Weinstein, Misha Seltzer"

import cgi
import datetime
import types


class DataTableException(Exception):
  """The general exception object thrown by DataTable."""
  pass


class DataTable(object):

  """Wraps the data to convert to a Google Visualization API DataTable.

  Create this object, populate it with data, then call one of the ToJS...
  methods to return a string representation of the data in the format described.

  You can clear all data from the object to reuse it, but you cannot clear
  individual cells, rows, or columns. You also cannot modify the table schema
  specified in the class constructor.

  You can add new data one or more rows at a time. All data added to an
  instantiated DataTable must conform to the schema passed in to __init__().

  You can reorder the columns in the output table, and also specify row sorting
  order by column. The default column order is according to the original
  table_description parameter. Default row sort order is ascending, by column
  1 values. For a dictionary, we sort the keys for order.

  The data and the table_description are closely tied, as described here:

  The table schema is defined in the class constructor's table_description
  parameter. The user defines each column using a tuple of
  (id[, type, [label]]). The default value for type is string, and label
  is the same as ID if not specified.

  table_description is a dictionary or list, containing one or more column
  descriptor tuples, nested dictionaries, and lists. Each dictionary key, list
  element, or dictionary element must eventually be defined as
  a column description tuple. Here's an example of a dictionary where the key
  is a tuple, and the value is a list of two tuples:
    {('a', 'number'): [('b', 'number'), ('c', 'string')]}

  This flexibility in data entry enables you to build and manipulate your data
  in a Python structure that makes sense for your program.

  Add data to the table using the same nested design as the table's
  table_description, replacing column descriptor tuples with cell data, and
  each row is an element in the top level collection. This will be a bit
  clearer after you look at the following examples showing the
  table_description, matching data, and the resulting table:

  Columns as list of tuples [col1, col2, col3]
    table_description: [('a', 'number'), ('b', 'string')]
    AppendData( [[1, 'z'], [2, 'w'], [4, 'o'], [5, 'k']] )
    Table:
    a  b   <--- these are column ids/labels
    1  z
    2  w
    4  o
    5  k

  Dictionary of columns, where key is a column, and value is a list of
  columns  {col1: [col2, col3]}
    table_description: {('a', 'number'): [('b', 'number'), ('c', 'string')]}
    AppendData( data: {1: [2, 'z'], 3: [4, 'w']}
    Table:
    a  b  c
    1  2  z
    3  4  w

  Dictionary where key is a column, and the value is itself a dictionary of
  columns {col1: {col2, col3}}
    table_description: {('a', 'number'): {'b': 'number', 'c': 'string'}}
    AppendData( data: {1: {'b': 2, 'c': 'z'}, 3: {'b': 4, 'c': 'w'}}
    Table:
    a  b  c
    1  2  z
    3  4  w
  """

  def __init__(self, table_description, data=None):
    """Initialize the data table from a table schema and (optionally) data.

    See the class documentation for more information on table schema and data
    values.

    Args:
      table_description: A table schema, following one of the formats described
                         in TableDescriptionParser(). Schemas describe the
                         column names, data types, and labels. See
                         TableDescriptionParser() for acceptable formats.
      data: Optional. If given, fills the table with the given data. The data
            structure must be consistent with schema in table_description. See
            the class documentation for more information on acceptable data. You
            can add data later by calling AppendData().

    Raises:
      DataTableException: Raised if the data and the description did not match,
                          or did not use the supported formats.
    """
    self.__columns = self.TableDescriptionParser(table_description)
    self.__data = []
    if data:
      self.LoadData(data)

  @staticmethod
  def _EscapeValue(v):
    """Puts the string in quotes, and escapes any inner quotes and slashes."""
    if isinstance(v, unicode):
      # Here we use repr as in the usual case, but on unicode strings, it
      # also escapes the unicode characters (which we want to leave as is).
      # So, after repr() we decode using raw-unicode-escape, which decodes
      # only the unicode characters, and leaves all the rest (", ', \n and
      # more) escaped.
      # We don't take the first character, because repr adds a u in the
      # beginning of the string (usual repr output for unicode is u'...').
      return repr(v).decode("raw-unicode-escape")[1:]
    # Here we use python built-in escaping mechanism for string using repr.
    return repr(str(v))

  @staticmethod
  def SingleValueToJS(value, value_type):
    """Translates a single value and type into a JS value.

    Internal helper method.

    Args:
      value: The value which should be converted
      value_type: One of "string", "number", "boolean", "date", "datetime" or
                  "timeofday".

    Returns:
      The proper JS format (as string) of the given value according to the
      given value_type. For None, we simply return "null".
      If a tuple is given, it should be of the form (value, formatted value)
      where the formatted value is a string. In such a case, we return
      the tuple of (JS value as string, formatted value).
      The real type of the given value is not strictly checked. For example,
      any type can be used for string - as we simply take its str( ) and for
      boolean value we just check "if value".
      Examples:
        SingleValueToJS(False, "boolean") returns "false"
        SingleValueToJS((5, "5$"), "number") returns ("5", "'5$'")

    Raises:
      DataTableException: The value and type did not match in a not-recoverable
                          way, for example given value 'abc' for type 'number'.
    """
    if isinstance(value, tuple):
      # In case of a tuple, we run the same function on the value itself and
      # add the formatted value.
      if len(value) != 2:
        raise DataTableException("Wrong format for value and formatting - %s." %
                                 str(value))
      if not isinstance(value[1], types.StringTypes):
        raise DataTableException("Formatted value is not string, given %s." %
                                 type(value[1]))
      js_value = DataTable.SingleValueToJS(value[0], value_type)
      if js_value == "null":
        raise DataTableException("An empty cell can not have formatting.")
      return (js_value, DataTable._EscapeValue(value[1]))

    # The standard case - no formatting.
    t_value = type(value)
    if value is None:
      return "null"
    if value_type == "boolean":
      if value:
        return "true"
      return "false"

    elif value_type == "number":
      if isinstance(value, (int, long, float)):
        return str(value)
      raise DataTableException("Wrong type %s when expected number" % t_value)

    elif value_type == "string":
      if isinstance(value, tuple):
        raise DataTableException("Tuple is not allowed as string value.")
      return DataTable._EscapeValue(value)

    elif value_type == "date":
      if not isinstance(value, (datetime.date, datetime.datetime)):
        raise DataTableException("Wrong type %s when expected date" % t_value)
        # We need to shift the month by 1 to match JS Date format
      return "new Date(%d,%d,%d)" % (value.year, value.month - 1, value.day)

    elif value_type == "timeofday":
      if not isinstance(value, (datetime.time, datetime.datetime)):
        raise DataTableException("Wrong type %s when expected time" % t_value)
      return "[%d,%d,%d]" % (value.hour, value.minute, value.second)

    elif value_type == "datetime":
      if not isinstance(value, datetime.datetime):
        raise DataTableException("Wrong type %s when expected datetime" %
                                 t_value)
      return "new Date(%d,%d,%d,%d,%d,%d)" % (value.year,
                                              value.month - 1,  # To match JS
                                              value.day,
                                              value.hour,
                                              value.minute,
                                              value.second)
    # If we got here, it means the given value_type was not one of the
    # supported types.
    raise DataTableException("Unsupported type %s" % value_type)

  @staticmethod
  def ColumnTypeParser(description):
    """Parses a single column description. Internal helper method.

    Args:
      description: a column description in the possible formats:
       'id'
       ('id',)
       ('id', 'type')
       ('id', 'type', 'label')
    Returns:
      Dictionary with the following keys: id, label and type where:
        - If label not given, it equals the id.
        - If type not given, string is used by default.

    Raises:
      DataTableException: The column description did not match the RE.
    """
    if not description:
      raise DataTableException("Description error: empty description given")

    if not isinstance(description, (types.StringTypes, tuple)):
      raise DataTableException("Description error: expected either string or "
                               "tuple, got %s." % type(description))

    if isinstance(description, types.StringTypes):
      description = (description,)

    # According to the tuple's length, we fill the keys
    # We verify everything is of type string
    for elem in description:
      if not isinstance(elem, types.StringTypes):
        raise DataTableException(("Description error: expected tuple of "
                                  "strings, current element of type %s." %
                                  type(elem)))
    desc_dict = {"id": description[0],
                 "label": description[0],
                 "type": "string"}
    if len(description) > 1:
      desc_dict["type"] = description[1].lower()
      if len(description) > 2:
        desc_dict["label"] = description[2]
        if len(description) > 3:
          raise DataTableException("Description error: tuple of length > 3")
    return desc_dict

  @staticmethod
  def TableDescriptionParser(table_description, depth=0):
    """Parses the table_description object for internal use.

    Parses the user-submitted table description into an internal format used
    by the Python DataTable class. Returns the flat list of parsed columns.

    Args:
      table_description: A description of the table which should comply
                         with one of the formats described below.
      depth: Optional. The depth of the first level in the current description.
             Used by recursive calls to this function.

    Returns:
      List of columns, where each column represented by a dictionary with the
      keys: id, label, type, depth, container which means the following:
      - id: the id of he column
      - name: The name of the column
      - type: The datatype of the elements in this column. Allowed types are
              described in ColumnTypeParser().
      - depth: The depth of this column in the table description
      - container: 'dict', 'iter' or 'scalar' for parsing the format easily.
      The returned description is flattened regardless of how it was given.

    Raises:
      DataTableException: Error in a column description or in the description
                          structure.

    Examples:
      A column description can be of the following forms:
       'id'
       ('id',)
       ('id', 'type')
       ('id', 'type', 'label')
       or as a dictionary:
       'id': 'type'
       'id': ('type',)
       'id': ('type', 'label')
      If the type is not specified, we treat it as string.
      If no specific label is given, the label is simply the id.

      input: [('a', 'date'), ('b', 'timeofday')]
      output: [{'id': 'a', 'label': 'a', 'type': 'date',
                'depth': 0, 'container': 'iter'},
               {'id': 'b', 'label': 'b', 'type': 'timeofday',
               'depth': 0, 'container': 'iter'}]

      input: {'a': [('b', 'number'), ('c', 'string', 'column c')]}
      output: [{'id': 'a', 'label': 'a', 'type': 'string',
                'depth': 0, 'container': 'dict'},
               {'id': 'b', 'label': 'b', 'type': 'number',
                'depth': 1, 'container': 'iter'},
               {'id': 'c', 'label': 'column c', 'type': 'string',
                'depth': 1, 'container': 'iter'}]

      input:  {('a', 'number', 'column a'): { 'b': 'number', 'c': 'string'}}
      output: [{'id': 'a', 'label': 'column a', 'type': 'number',
                'depth': 0, 'container': 'dict'},
               {'id': 'b', 'label': 'b', 'type': 'number',
                'depth': 1, 'container': 'dict'},
               {'id': 'c', 'label': 'c', 'type': 'string',
                'depth': 1, 'container': 'dict'}]

      input: { ('w', 'string', 'word'): ('c', 'number', 'count') }
      output: [{'id': 'w', 'label': 'word', 'type': 'string',
                'depth': 0, 'container': 'dict'},
               {'id': 'c', 'label': 'count', 'type': 'number',
                'depth': 1, 'container': 'scalar'}]
    """
    # For the recursion step, we check for a scalar object (string or tuple)
    if isinstance(table_description, (types.StringTypes, tuple)):
      parsed_col = DataTable.ColumnTypeParser(table_description)
      parsed_col["depth"] = depth
      parsed_col["container"] = "scalar"
      return [parsed_col]

    # Since it is not scalar, table_description must be iterable.
    if not hasattr(table_description, "__iter__"):
      raise DataTableException("Expected an iterable object, got %s" %
                               type(table_description))
    if not isinstance(table_description, dict):
      # We expects a non-dictionary iterable item.
      columns = []
      for desc in table_description:
        parsed_col = DataTable.ColumnTypeParser(desc)
        parsed_col["depth"] = depth
        parsed_col["container"] = "iter"
        columns.append(parsed_col)
      if not columns:
        raise DataTableException("Description iterable objects should not"
                                 " be empty.")
      return columns
    # The other case is a dictionary
    if not table_description:
      raise DataTableException("Empty dictionaries are not allowed inside"
                               " description")

    # The number of keys in the dictionary separates between the two cases of
    # more levels below or this is the most inner dictionary.
    if len(table_description) != 1:
      # This is the most inner dictionary. Parsing types.
      columns = []
      # We sort the items, equivalent to sort the keys since they are unique
      for key, value in sorted(table_description.items()):
        # We parse the column type as (key, type) or (key, type, label) using
        # ColumnTypeParser.
        if isinstance(value, tuple):
          parsed_col = DataTable.ColumnTypeParser((key,) + value)
        else:
          parsed_col = DataTable.ColumnTypeParser((key, value))
        parsed_col["depth"] = depth
        parsed_col["container"] = "dict"
        columns.append(parsed_col)
      return columns
    # This is an outer dictionary, must have at most one key.
    parsed_col = DataTable.ColumnTypeParser(table_description.keys()[0])
    parsed_col["depth"] = depth
    parsed_col["container"] = "dict"
    return ([parsed_col] +
            DataTable.TableDescriptionParser(table_description.values()[0],
                                             depth=depth + 1))

  @property
  def columns(self):
    """Returns the parsed table description."""
    return self.__columns

  def NumberOfRows(self):
    """Returns the number of rows in the current data stored in the table."""
    return len(self.__data)

  def LoadData(self, data):
    """Loads new data to the data table, clearing existing data."""
    self.__data = []
    self.AppendData(data)

  def AppendData(self, data):
    """Appends new data to the table.

    Data is appended in rows. Data must comply with
    the table schema passed in to __init__(). See SingleValueToJS() for a list
    of acceptable data types. See the class documentation for more information
    and examples of schema and data values.

    Args:
      data: The row to add to the table. The data must conform to the table
            description format.

    Raises:
      DataTableException: The data structure does not match the description.
    """
    # If the maximal depth is 0, we simply iterate over the data table
    # lines and insert them using _InnerAppendData. Otherwise, we simply
    # let the _InnerAppendData handle all the levels.
    if not self.__columns[-1]["depth"]:
      for line in data:
        self._InnerAppendData({}, line, 0)
    else:
      self._InnerAppendData({}, data, 0)

  def _InnerAppendData(self, prev_col_values, data, col_index):
    """Inner function to assist LoadData."""
    # We first check that col_index has not exceeded the columns size
    if col_index >= len(self.__columns):
      raise DataTableException("The data does not match description, too deep")

    # Dealing with the scalar case, the data is the last value.
    if self.__columns[col_index]["container"] == "scalar":
      prev_col_values[self.__columns[col_index]["id"]] = data
      self.__data.append(prev_col_values)
      return

    if self.__columns[col_index]["container"] == "iter":
      if not hasattr(data, "__iter__") or isinstance(data, dict):
        raise DataTableException("Expected iterable object, got %s" %
                                 type(data))
      # We only need to insert the rest of the columns
      # If there are less items than expected, we only add what there is.
      for value in data:
        if col_index >= len(self.__columns):
          raise DataTableException("Too many elements given in data")
        prev_col_values[self.__columns[col_index]["id"]] = value
        col_index += 1
      self.__data.append(prev_col_values)
      return

    # We know the current level is a dictionary, we verify the type.
    if not isinstance(data, dict):
      raise DataTableException("Expected dictionary at current level, got %s" %
                               type(data))
    # We check if this is the last level
    if self.__columns[col_index]["depth"] == self.__columns[-1]["depth"]:
      # We need to add the keys in the dictionary as they are
      for col in self.__columns[col_index:]:
        if col["id"] in data:
          prev_col_values[col["id"]] = data[col["id"]]
      self.__data.append(prev_col_values)
      return

    # We have a dictionary in an inner depth level.
    if not data.keys():
      # In case this is an empty dictionary, we add a record with the columns
      # filled only until this point.
      self.__data.append(prev_col_values)
    else:
      for key in sorted(data):
        col_values = dict(prev_col_values)
        col_values[self.__columns[col_index]["id"]] = key
        self._InnerAppendData(col_values, data[key], col_index + 1)

  def _PreparedData(self, order_by=()):
    """Prepares the data for enumeration - sorting it by order_by.

    Args:
      order_by: Optional. Specifies the name of the column(s) to sort by, and
                (optionally) which direction to sort in. Default sort direction
                is asc. Following formats are accepted:
                "string_col_name"  -- For a single key in default (asc) order.
                ("string_col_name", "asc|desc") -- For a single key.
                [("col_1","asc|desc"), ("col_2","asc|desc")] -- For more than
                    one column, an array of tuples of (col_name, "asc|desc").

    Returns:
      The data sorted by the keys given.

    Raises:
      DataTableException: Sort direction not in 'asc' or 'desc'
    """
    if not order_by:
      return self.__data

    proper_sort_keys = []
    if isinstance(order_by, types.StringTypes) or (
        isinstance(order_by, tuple) and len(order_by) == 2 and
        order_by[1].lower() in ["asc", "desc"]):
      order_by = (order_by,)
    for key in order_by:
      if isinstance(key, types.StringTypes):
        proper_sort_keys.append((key, 1))
      elif (isinstance(key, (list, tuple)) and len(key) == 2 and
            key[1].lower() in ("asc", "desc")):
        proper_sort_keys.append((key[0], key[1].lower() == "asc" and 1 or -1))
      else:
        raise DataTableException("Expected tuple with second value: "
                                 "'asc' or 'desc'")

    def SortCmpFunc(row1, row2):
      """cmp function for sorted. Compares by keys and 'asc'/'desc' keywords."""
      for key, asc_mult in proper_sort_keys:
        cmp_result = asc_mult * cmp(row1.get(key), row2.get(key))
        if cmp_result:
          return cmp_result
      return 0

    return sorted(self.__data, cmp=SortCmpFunc)

  def ToJSCode(self, name, columns_order=None, order_by=()):
    """Writes the data table as a JS code string.

    This method writes a string of JS code that can be run to
    generate a DataTable with the specified data. Typically used for debugging
    only.

    Args:
      name: The name of the table. The name would be used as the DataTable's
            variable name in the created JS code.
      columns_order: Optional. Specifies the order of columns in the
                     output table. Specify a list of all column IDs in the order
                     in which you want the table created.
                     Note that you must list all column IDs in this parameter,
                     if you use it.
      order_by: Optional. Specifies the name of the column(s) to sort by.
                Passed as is to _PreparedData.

    Returns:
      A string of JS code that, when run, generates a DataTable with the given
      name and the data stored in the DataTable object.
      Example result:
        "var tab1 = new google.visualization.DataTable();
         tab1.addColumn('string', 'a', 'a');
         tab1.addColumn('number', 'b', 'b');
         tab1.addColumn('boolean', 'c', 'c');
         tab1.addRows(10);
         tab1.setCell(0, 0, 'a');
         tab1.setCell(0, 1, 1);
         tab1.setCell(0, 2, true);
         ...
         tab1.setCell(9, 0, 'c');
         tab1.setCell(9, 1, 3, '3$');
         tab1.setCell(9, 2, false);"

    Raises:
      DataTableException: The data does not match the type.
    """
    if columns_order is None:
      columns_order = [col["id"] for col in self.__columns]
    col_dict = dict([(col["id"], col) for col in self.__columns])

    # We first create the table with the given name
    jscode = "var %s = new google.visualization.DataTable();\n" % name

    # We add the columns to the table
    for col in columns_order:
      jscode += "%s.addColumn('%s', '%s', '%s');\n" % (name,
                                                       col_dict[col]["type"],
                                                       col_dict[col]["label"],
                                                       col_dict[col]["id"])
    jscode += "%s.addRows(%d);\n" % (name, len(self.__data))

    # We now go over the data and add each row
    for (i, row) in enumerate(self._PreparedData(order_by)):
      # We add all the elements of this row by their order
      for (j, col) in enumerate(columns_order):
        if col not in row or row[col] is None:
          continue
        value = self.SingleValueToJS(row[col], col_dict[col]["type"])
        if isinstance(value, tuple):
          # We have a formatted value as well
          jscode += ("%s.setCell(%d, %d, %s, %s);\n" %
                     (name, i, j, value[0], value[1]))
        else:
          jscode += "%s.setCell(%d, %d, %s);\n" % (name, i, j, value)
    return jscode

  def ToHtml(self, columns_order=None, order_by=()):
    """Writes the data table as an HTML table code string.

    Args:
      columns_order: Optional. Specifies the order of columns in the
                     output table. Specify a list of all column IDs in the order
                     in which you want the table created.
                     Note that you must list all column IDs in this parameter,
                     if you use it.
      order_by: Optional. Specifies the name of the column(s) to sort by.
                Passed as is to _PreparedData.

    Returns:
      An HTML table code string.
      Example result (the result is without the newlines):
       <html><body><table border='1'>
        <thead><tr><th>a</th><th>b</th><th>c</th></tr></thead>
        <tbody>
         <tr><td>1</td><td>"z"</td><td>2</td></tr>
         <tr><td>"3$"</td><td>"w"</td><td></td></tr>
        </tbody>
       </table></body></html>

    Raises:
      DataTableException: The data does not match the type.
    """
    table_template = "<html><body><table border='1'>%s</table></body></html>"
    columns_template = "<thead><tr>%s</tr></thead>"
    rows_template = "<tbody>%s</tbody>"
    row_template = "<tr>%s</tr>"
    header_cell_template = "<th>%s</th>"
    cell_template = "<td>%s</td>"

    if columns_order is None:
      columns_order = [col["id"] for col in self.__columns]
    col_dict = dict([(col["id"], col) for col in self.__columns])

    columns_list = []
    for col in columns_order:
      columns_list.append(header_cell_template % col_dict[col]["label"])
    columns_html = columns_template % "".join(columns_list)

    rows_list = []
    # We now go over the data and add each row
    for row in self._PreparedData(order_by):
      cells_list = []
      # We add all the elements of this row by their order
      for col in columns_order:
        # For empty string we want empty quotes ("").
        value = ""
        if col in row and row[col] is not None:
          value = self.SingleValueToJS(row[col], col_dict[col]["type"])
        if isinstance(value, tuple):
          # We have a formatted value and we're going to use it
          cells_list.append(cell_template % cgi.escape(value[1]))
        else:
          cells_list.append(cell_template % cgi.escape(value))
      rows_list.append(row_template % "".join(cells_list))
    rows_html = rows_template % "".join(rows_list)

    return table_template % (columns_html + rows_html)

  def ToCsv(self, columns_order=None, order_by=()):
    """Writes the data table as a CSV string.

    Args:
      columns_order: Optional. Specifies the order of columns in the
                     output table. Specify a list of all column IDs in the order
                     in which you want the table created.
                     Note that you must list all column IDs in this parameter,
                     if you use it.
      order_by: Optional. Specifies the name of the column(s) to sort by.
                Passed as is to _PreparedData.

    Returns:
      A CSV string representing the table.
      Example result:
       'a', 'b', 'c'
       1, 'z', 2
       3, 'w', ''

    Raises:
      DataTableException: The data does not match the type.
    """
    if columns_order is None:
      columns_order = [col["id"] for col in self.__columns]
    col_dict = dict([(col["id"], col) for col in self.__columns])

    columns_list = []
    for col in columns_order:
      columns_list.append(DataTable._EscapeValue(col_dict[col]["label"]))
    columns_line = ", ".join(columns_list)

    rows_list = []
    # We now go over the data and add each row
    for row in self._PreparedData(order_by):
      cells_list = []
      # We add all the elements of this row by their order
      for col in columns_order:
        value = "''"
        if col in row and row[col] is not None:
          value = self.SingleValueToJS(row[col], col_dict[col]["type"])
        if isinstance(value, tuple):
          # We have a formatted value. Using it only for date/time types.
          if col_dict[col]["type"] in ["date", "datetime", "timeofday"]:
            cells_list.append(value[1])
          else:
            cells_list.append(value[0])
        else:
          # We need to quote date types, because they contain commas.
          if (col_dict[col]["type"] in ["date", "datetime", "timeofday"] and
              value != "''"):
            value = "'%s'" % value
          cells_list.append(value)
      rows_list.append(", ".join(cells_list))
    rows = "\n".join(rows_list)

    return "%s\n%s" % (columns_line, rows)

  def ToJSon(self, columns_order=None, order_by=()):
    """Writes a JSON string that can be used in a JS DataTable constructor.

    This method writes a JSON string that can be passed directly into a Google
    Visualization API DataTable constructor. Use this output if you are
    hosting the visualization HTML on your site, and want to code the data
    table in Python. Pass this string into the
    google.visualization.DataTable constructor, e.g,:
      ... on my page that hosts my visualization ...
      google.setOnLoadCallback(drawTable);
      function drawTable() {
        var data = new google.visualization.DataTable(_my_JSon_string, 0.5);
        myTable.draw(data);
      }

    Args:
      columns_order: Optional. Specifies the order of columns in the
                     output table. Specify a list of all column IDs in the order
                     in which you want the table created.
                     Note that you must list all column IDs in this parameter,
                     if you use it.
      order_by: Optional. Specifies the name of the column(s) to sort by.
                Passed as is to _PreparedData().

    Returns:
      A JSon constructor string to generate a JS DataTable with the data
      stored in the DataTable object.
      Example result (the result is without the newlines):
       {cols: [{id:'a',label:'a',type:'number'},
               {id:'b',label:'b',type:'string'},
              {id:'c',label:'c',type:'number'}],
        rows: [{c:[{v:1},{v:'z'},{v:2}]}, c:{[{v:3,f:'3$'},{v:'w'},{v:null}]}]}

    Raises:
      DataTableException: The data does not match the type.
    """
    if columns_order is None:
      columns_order = [col["id"] for col in self.__columns]
    col_dict = dict([(col["id"], col) for col in self.__columns])

    # Creating the columns jsons
    cols_jsons = ["{id:'%(id)s',label:'%(label)s',type:'%(type)s'}" %
                  col_dict[col_id] for col_id in columns_order]

    # Creating the rows jsons
    rows_jsons = []
    for row in self._PreparedData(order_by):
      cells_jsons = []
      for col in columns_order:
        # We omit the {v:null} for a None value of the not last column
        value = row.get(col, None)
        if value is None and col != columns_order[-1]:
          cells_jsons.append("")
        else:
          value = self.SingleValueToJS(value, col_dict[col]["type"])
          if isinstance(value, tuple):
            # We have a formatted value as well
            cells_jsons.append("{v:%s,f:%s}" % value)
          else:
            cells_jsons.append("{v:%s}" % value)
      rows_jsons.append("{c:[%s]}" % ",".join(cells_jsons))

    # We now join the columns jsons and the rows jsons
    json = "{cols: [%s],rows: [%s]}" % (",".join(cols_jsons),
                                        ",".join(rows_jsons))
    return json

  def ToJSonResponse(self, columns_order=None, order_by=(), req_id=0,
                     response_handler="google.visualization.Query.setResponse"):
    """Writes a table as a JSON response that can be returned as-is to a client.

    This method writes a JSON response to return to a client in response to a
    Google Visualization API query. This string can be processed by the calling
    page, and is used to deliver a data table to a visualization hosted on
    a different page.

    Args:
      columns_order: Optional. Passed straight to self.ToJSon().
      order_by: Optional. Passed straight to self.ToJSon().
      req_id: Optional. The response id, as retrieved by the request.
      response_handler: Optional. The response handler, as retrieved by the
          request.

    Returns:
      A JSON response string to be received by JS the visualization Query
      object. This response would be translated into a DataTable on the
      client side.
      Example result (newlines added for readability):
       google.visualization.Query.setResponse({
          'version':'0.5', 'reqId':'0', 'status':'OK',
          'table': {cols: [...], rows: [...]}});

    Note: The URL returning this string can be used as a data source by Google
          Visualization Gadgets or from JS code.
    """
    table = self.ToJSon(columns_order, order_by)
    return ("%s({'version':'0.5', 'reqId':'%s', 'status':'OK', "
            "'table': %s});") % (response_handler, req_id, table)

  def ToResponse(self, columns_order=None, order_by=(), tqx=""):
    """Writes the right response according to the request string passed in tqx.

    This method parses the tqx request string (format of which is defined in
    the documentation for implementing a data source of Google Visualization),
    and returns the right response according to the request.
    It parses out the "out" parameter of tqx, calls the relevant response
    (ToJSonResponse() for "json", ToCsv() for "csv" and ToHtml() for "html")
    and passes the response function the rest of the relevant request keys.

    Args:
      columns_order: Optional. Passed as is to the relevant response function.
      order_by: Optional. Passed as is to the relevant response function.
      tqx: Optional. The request string as received by HTTP GET. Should be in
           the format "key1:value1;key2:value2...". All keys have a default
           value, so an empty string will just do the default (which is calling
           ToJSonResponse() with no extra parameters).

    Returns:
      A response string, as returned by the relevant response function.

    Raises:
      DataTableException: One of the parameters passed in tqx is not supported.
    """
    tqx_dict = {}
    if tqx:
      tqx_dict = dict(opt.split(":") for opt in tqx.split(";"))
    if tqx_dict.get("version", 0.5) != 0.5:
      raise DataTableException(
          "Version (%s) passed by request is not supported."
          % tqx_dict["version"])

    if tqx_dict.get("out", "json") == "json":
      response_handler = tqx_dict.get("responseHandler",
                                      "google.visualization.Query.setResponse")
      return self.ToJSonResponse(columns_order, order_by,
                                 req_id=tqx_dict.get("reqId", 0),
                                 response_handler=response_handler)
    elif tqx_dict["out"] == "html":
      return self.ToHtml(columns_order, order_by)
    elif tqx_dict["out"] == "csv":
      return self.ToCsv(columns_order, order_by)
    else:
      raise DataTableException(
          "'out' parameter: '%s' is not supported" % tqx_dict["out"])
